"""
FocalNet, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/focalnet.py

Paper "Focal Modulation Networks", https://arxiv.org/abs/2203.11926
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from collections.abc import Callable
from functools import partial
from typing import Any
from typing import Literal
from typing import Optional

import torch
from torch import nn
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.layers import LayerNorm2d
from birder.layers import LayerScale2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType


class MLP(nn.Module):
    def __init__(
        self,
        in_features: int,
        hidden_features: int,
        out_features: int,
        act_layer: Callable[..., nn.Module],
        drop: float,
    ) -> None:
        super().__init__()
        self.fc1 = nn.Conv2d(in_features, hidden_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=True)
        self.act = act_layer()
        self.drop1 = nn.Dropout(drop)
        self.fc2 = nn.Conv2d(
            hidden_features, out_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=True
        )
        self.drop2 = nn.Dropout(drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop1(x)
        x = self.fc2(x)
        x = self.drop2(x)

        return x


class Downsample(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int = 4,
        overlap: bool = False,
        norm_layer: Callable[..., torch.Tensor] = LayerNorm2d,
    ) -> None:
        super().__init__()
        padding = (0, 0)
        kernel_size = (stride, stride)
        if overlap is True:
            assert stride in (2, 4)
            if stride == 4:
                kernel_size = (7, 7)
                padding = (2, 2)

            elif stride == 2:
                kernel_size = (3, 3)
                padding = (1, 1)

        self.proj = nn.Conv2d(
            in_channels, out_channels, kernel_size=kernel_size, stride=stride, padding=padding, bias=True
        )
        self.norm = norm_layer(out_channels) if norm_layer is not None else nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.proj(x)
        x = self.norm(x)

        return x


class FocalModulation(nn.Module):
    def __init__(
        self,
        dim: int,
        focal_window: int,
        focal_level: int,
        focal_factor: int,
        bias: bool,
        use_post_norm: bool,
        normalize_modulator: bool,
        proj_drop: float,
        norm_layer: Callable[..., torch.Tensor] = LayerNorm2d,
    ) -> None:
        super().__init__()

        self.dim = dim
        self.focal_window = focal_window
        self.focal_level = focal_level
        self.focal_factor = focal_factor
        self.use_post_norm = use_post_norm
        self.normalize_modulator = normalize_modulator
        self.input_split = [dim, dim, self.focal_level + 1]

        self.f = nn.Conv2d(
            dim,
            2 * dim + (self.focal_level + 1),
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            bias=bias,
        )
        self.h = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=bias)

        self.act = nn.GELU()
        self.proj = nn.Conv2d(dim, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=True)
        self.proj_drop = nn.Dropout(proj_drop)
        self.focal_layers = nn.ModuleList()

        self.kernel_sizes = []
        for k in range(self.focal_level):
            kernel_size = self.focal_factor * k + self.focal_window
            self.focal_layers.append(
                nn.Sequential(
                    nn.Conv2d(
                        dim,
                        dim,
                        kernel_size=kernel_size,
                        stride=(1, 1),
                        padding=kernel_size // 2,
                        groups=dim,
                        bias=False,
                    ),
                    nn.GELU(),
                )
            )
            self.kernel_sizes.append(kernel_size)

        self.norm = norm_layer(dim) if self.use_post_norm else nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Pre linear projection
        x = self.f(x)
        (q, ctx, gates) = torch.split(x, self.input_split, 1)

        # Context aggregation
        ctx_all = 0.0
        for idx, focal_layer in enumerate(self.focal_layers):
            ctx = focal_layer(ctx)
            ctx_all = ctx_all + ctx * gates[:, idx : idx + 1]

        ctx_global = self.act(ctx.mean((2, 3), keepdim=True))
        ctx_all = ctx_all + ctx_global * gates[:, self.focal_level :]

        # Normalize context
        if self.normalize_modulator is True:
            ctx_all = ctx_all / (self.focal_level + 1)

        # Focal modulation
        x_out = q * self.h(ctx_all)
        x_out = self.norm(x_out)

        # Post linear projection
        x_out = self.proj(x_out)
        x_out = self.proj_drop(x_out)

        return x_out


class FocalNetBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        mlp_ratio: float,
        focal_level: int,
        focal_window: int,
        use_post_norm: bool,
        use_post_norm_in_modulation: bool,
        normalize_modulator: bool,
        layer_scale_value: Optional[float],
        proj_drop: float,
        drop_path: float,
        norm_layer: Callable[..., torch.Tensor] = LayerNorm2d,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.mlp_ratio = mlp_ratio

        self.focal_window = focal_window
        self.focal_level = focal_level
        self.use_post_norm = use_post_norm

        self.norm1 = norm_layer(dim) if not use_post_norm else nn.Identity()
        self.modulation = FocalModulation(
            dim,
            focal_window=focal_window,
            focal_level=self.focal_level,
            focal_factor=2,
            bias=True,
            use_post_norm=use_post_norm_in_modulation,
            normalize_modulator=normalize_modulator,
            proj_drop=proj_drop,
            norm_layer=norm_layer,
        )
        self.norm1_post = norm_layer(dim) if use_post_norm else nn.Identity()
        self.ls1 = LayerScale2d(dim, layer_scale_value) if layer_scale_value is not None else nn.Identity()
        self.drop_path1 = StochasticDepth(drop_path, mode="row")

        self.norm2 = norm_layer(dim) if not use_post_norm else nn.Identity()
        self.mlp = MLP(
            in_features=dim,
            hidden_features=int(dim * mlp_ratio),
            out_features=dim,
            act_layer=nn.GELU,
            drop=proj_drop,
        )
        self.norm2_post = norm_layer(dim) if use_post_norm else nn.Identity()
        self.ls2 = LayerScale2d(dim, layer_scale_value) if layer_scale_value is not None else nn.Identity()
        self.drop_path2 = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        shortcut = x

        # Focal modulation
        x = self.norm1(x)
        x = self.modulation(x)
        x = self.norm1_post(x)
        x = shortcut + self.drop_path1(self.ls1(x))

        # FFN
        x = x + self.drop_path2(self.ls2(self.norm2_post(self.mlp(self.norm2(x)))))

        return x


class FocalNetStage(nn.Module):
    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(
        self,
        dim: int,
        out_dim: int,
        depth: int,
        mlp_ratio: float,
        downsample: bool,
        focal_level: int,
        focal_window: int,
        use_overlap_down: bool,
        use_post_norm: bool,
        use_post_norm_in_modulation: bool,
        normalize_modulator: bool,
        layer_scale_value: Optional[float],
        proj_drop: float,
        drop_path: list[float],
        norm_layer: Callable[..., torch.Tensor] = LayerNorm2d,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.depth = depth
        if downsample is True:
            self.downsample = Downsample(
                in_channels=dim,
                out_channels=out_dim,
                stride=2,
                overlap=use_overlap_down,
                norm_layer=norm_layer,
            )

        else:
            self.downsample = nn.Identity()

        blocks = []
        for i in range(depth):
            blocks.append(
                FocalNetBlock(
                    dim=out_dim,
                    mlp_ratio=mlp_ratio,
                    focal_level=focal_level,
                    focal_window=focal_window,
                    use_post_norm=use_post_norm,
                    use_post_norm_in_modulation=use_post_norm_in_modulation,
                    normalize_modulator=normalize_modulator,
                    layer_scale_value=layer_scale_value,
                    proj_drop=proj_drop,
                    drop_path=drop_path[i],
                    norm_layer=norm_layer,
                )
            )

        self.blocks = nn.Sequential(*blocks)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)

        return x


class FocalNet(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.blocks\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        proj_drop_rate = 0.0
        depths: list[int] = self.config["depths"]
        embed_dim: int = self.config["embed_dim"]
        focal_levels: tuple[int, int, int, int] = self.config["focal_levels"]
        focal_windows: tuple[int, int, int, int] = self.config["focal_windows"]
        layer_scale_value: Optional[float] = self.config["layer_scale_value"]
        use_post_norm: bool = self.config["use_post_norm"]
        use_overlap_down: bool = self.config["use_overlap_down"]
        use_post_norm_in_modulation: bool = self.config["use_post_norm_in_modulation"]
        drop_path_rate: float = self.config["drop_path_rate"]

        num_stages = len(depths)
        embed_dims = [embed_dim * (2**i) for i in range(num_stages)]
        num_features = embed_dims[-1]
        norm_layer = partial(LayerNorm2d, eps=1e-5)

        self.stem = Downsample(
            in_channels=3,
            out_channels=embed_dims[0],
            overlap=use_overlap_down,
            norm_layer=norm_layer,
        )

        in_dim = embed_dims[0]
        dpr: list[float] = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for idx in range(num_stages):
            out_dim = embed_dims[idx]
            stages[f"stage{idx+1}"] = FocalNetStage(
                dim=in_dim,
                out_dim=out_dim,
                depth=depths[idx],
                mlp_ratio=4.0,
                downsample=idx > 0,
                focal_level=focal_levels[idx],
                focal_window=focal_windows[idx],
                use_overlap_down=use_overlap_down,
                use_post_norm=use_post_norm,
                use_post_norm_in_modulation=use_post_norm_in_modulation,
                normalize_modulator=False,
                layer_scale_value=layer_scale_value,
                proj_drop=proj_drop_rate,
                drop_path=dpr[sum(depths[:idx]) : sum(depths[: idx + 1])],
                norm_layer=norm_layer,
            )

            return_channels.append(out_dim)
            in_dim = out_dim

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            norm_layer(num_features),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = num_features
        self.classifier = self.create_classifier()

        self.stem_stride = 4
        self.stem_width = embed_dims[0]
        self.encoding_size = num_features

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "focalnet_t_srf",
    FocalNet,
    config={
        "depths": [2, 2, 6, 2],
        "embed_dim": 96,
        "focal_levels": (2, 2, 2, 2),
        "focal_windows": (3, 3, 3, 3),
        "layer_scale_value": None,
        "use_post_norm": False,
        "use_overlap_down": False,
        "use_post_norm_in_modulation": False,
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "focalnet_t_lrf",
    FocalNet,
    config={
        "depths": [2, 2, 6, 2],
        "embed_dim": 96,
        "focal_levels": (3, 3, 3, 3),
        "focal_windows": (3, 3, 3, 3),
        "layer_scale_value": None,
        "use_post_norm": False,
        "use_overlap_down": False,
        "use_post_norm_in_modulation": False,
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "focalnet_s_srf",
    FocalNet,
    config={
        "depths": [2, 2, 18, 2],
        "embed_dim": 96,
        "focal_levels": (2, 2, 2, 2),
        "focal_windows": (3, 3, 3, 3),
        "layer_scale_value": None,
        "use_post_norm": False,
        "use_overlap_down": False,
        "use_post_norm_in_modulation": False,
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "focalnet_s_lrf",
    FocalNet,
    config={
        "depths": [2, 2, 18, 2],
        "embed_dim": 96,
        "focal_levels": (3, 3, 3, 3),
        "focal_windows": (3, 3, 3, 3),
        "layer_scale_value": None,
        "use_post_norm": False,
        "use_overlap_down": False,
        "use_post_norm_in_modulation": False,
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "focalnet_b_srf",
    FocalNet,
    config={
        "depths": [2, 2, 18, 2],
        "embed_dim": 128,
        "focal_levels": (2, 2, 2, 2),
        "focal_windows": (3, 3, 3, 3),
        "layer_scale_value": None,
        "use_post_norm": False,
        "use_overlap_down": False,
        "use_post_norm_in_modulation": False,
        "drop_path_rate": 0.3,
    },
)
registry.register_model_config(
    "focalnet_b_lrf",
    FocalNet,
    config={
        "depths": [2, 2, 18, 2],
        "embed_dim": 128,
        "focal_levels": (3, 3, 3, 3),
        "focal_windows": (3, 3, 3, 3),
        "layer_scale_value": None,
        "use_post_norm": False,
        "use_overlap_down": False,
        "use_post_norm_in_modulation": False,
        "drop_path_rate": 0.3,
    },
)
registry.register_model_config(
    "focalnet_l3",
    FocalNet,
    config={
        "depths": [2, 2, 18, 2],
        "embed_dim": 192,
        "focal_levels": (3, 3, 3, 3),
        "focal_windows": (5, 5, 5, 5),
        "layer_scale_value": 1e-4,
        "use_post_norm": True,
        "use_overlap_down": True,
        "use_post_norm_in_modulation": False,
        "drop_path_rate": 0.3,
    },
)
registry.register_model_config(
    "focalnet_l4",
    FocalNet,
    config={
        "depths": [2, 2, 18, 2],
        "embed_dim": 192,
        "focal_levels": (4, 4, 4, 4),
        "focal_windows": (3, 3, 3, 3),
        "layer_scale_value": 1e-4,
        "use_post_norm": True,
        "use_overlap_down": True,
        "use_post_norm_in_modulation": False,
        "drop_path_rate": 0.3,
    },
)
registry.register_model_config(
    "focalnet_xl3",
    FocalNet,
    config={
        "depths": [2, 2, 18, 2],
        "embed_dim": 256,
        "focal_levels": (3, 3, 3, 3),
        "focal_windows": (5, 5, 5, 5),
        "layer_scale_value": 1e-4,
        "use_post_norm": True,
        "use_overlap_down": True,
        "use_post_norm_in_modulation": False,
        "drop_path_rate": 0.3,
    },
)
registry.register_model_config(
    "focalnet_xl4",
    FocalNet,
    config={
        "depths": [2, 2, 18, 2],
        "embed_dim": 256,
        "focal_levels": (4, 4, 4, 4),
        "focal_windows": (3, 3, 3, 3),
        "layer_scale_value": 1e-4,
        "use_post_norm": True,
        "use_overlap_down": True,
        "use_post_norm_in_modulation": False,
        "drop_path_rate": 0.3,
    },
)
registry.register_model_config(
    "focalnet_h3",
    FocalNet,
    config={
        "depths": [2, 2, 18, 2],
        "embed_dim": 352,
        "focal_levels": (3, 3, 3, 3),
        "focal_windows": (5, 5, 5, 5),
        "layer_scale_value": 1e-4,
        "use_post_norm": True,
        "use_overlap_down": True,
        "use_post_norm_in_modulation": True,
        "drop_path_rate": 0.3,
    },
)
registry.register_model_config(
    "focalnet_h4",
    FocalNet,
    config={
        "depths": [2, 2, 18, 2],
        "embed_dim": 352,
        "focal_levels": (4, 4, 4, 4),
        "focal_windows": (3, 3, 3, 3),
        "layer_scale_value": 1e-4,
        "use_post_norm": True,
        "use_overlap_down": True,
        "use_post_norm_in_modulation": True,
        "drop_path_rate": 0.3,
    },
)

registry.register_weights(
    "focalnet_b_lrf_intermediate-eu-common",
    {
        "url": "https://huggingface.co/birder-project/focalnet_b_lrf_intermediate-eu-common/resolve/main",
        "description": (
            "FocalNet Base (LRF) model with intermediate training, then fine-tuned on the eu-common dataset"
        ),
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 337.6,
                "sha256": "9318e35c033606bb58d4992867d19fe4718e416d33955319a2841ed4ab7dcb5d",
            },
        },
        "net": {"network": "focalnet_b_lrf", "tag": "intermediate-eu-common"},
    },
)
registry.register_weights(
    "focalnet_b_lrf_intermediate-arabian-peninsula",
    {
        "url": "https://huggingface.co/birder-project/focalnet_b_lrf_intermediate-arabian-peninsula/resolve/main",
        "description": (
            "FocalNet Base (LRF) model with intermediate training, then fine-tuned on the arabian-peninsula dataset"
        ),
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 337.7,
                "sha256": "9a225c0a58cb4114ada7ea20c9410a4dea116670335cc0eabaa8448f8e9247ab",
            },
        },
        "net": {"network": "focalnet_b_lrf", "tag": "intermediate-arabian-peninsula"},
    },
)



import argparse
import logging
import os
import typing
from typing import Optional
from typing import get_args

from birder.common.cli import ValidationError
from birder.common.training_utils import OptimizerType
from birder.common.training_utils import SchedulerType
from birder.conf import settings
from birder.data.transforms.classification import AugType
from birder.data.transforms.classification import RGBMode
from birder.data.transforms.detection import AugType as DetAugType

logger = logging.getLogger(__name__)


def add_compile_args(parser: argparse.ArgumentParser, teacher: bool = False, backbone: bool = False) -> None:
    group = parser.add_argument_group("Compilation parameters")
    group.add_argument("--compile", default=False, action="store_true", help="enable compilation")
    if teacher is True:
        group.add_argument(
            "--compile-teacher", default=False, action="store_true", help="enable teacher only compilation"
        )
    if backbone is True:
        group.add_argument(
            "--compile-backbone", default=False, action="store_true", help="enable backbone only compilation"
        )

    group.add_argument(
        "--compile-opt", default=False, action="store_true", help="enable compilation for optimizer step"
    )


def add_optimization_args(parser: argparse.ArgumentParser, default_batch_size: int = 32) -> None:
    group = parser.add_argument_group("Optimization parameters")
    group.add_argument("--batch-size", type=int, default=default_batch_size, metavar="N", help="the batch size")
    group.add_argument("--opt", type=str, choices=list(get_args(OptimizerType)), default="sgd", help="optimizer to use")
    group.add_argument("--momentum", type=float, default=0.9, metavar="M", help="optimizer momentum")
    group.add_argument("--nesterov", default=False, action="store_true", help="use nesterov momentum")
    group.add_argument("--opt-eps", type=float, help="optimizer epsilon (None to use the optimizer default)")
    group.add_argument("--opt-betas", type=float, nargs="+", help="optimizer betas (None to use the optimizer default)")
    group.add_argument("--opt-alpha", type=float, help="optimizer alpha (None to use the optimizer default)")
    group.add_argument("--clip-grad-norm", type=float, help="the maximum gradient norm")
    group.add_argument(
        "--grad-accum-steps", type=int, default=1, metavar="N", help="number of steps to accumulate gradients"
    )


def add_lr_wd_args(parser: argparse.ArgumentParser, backbone_lr: bool = False, wd_end: bool = False) -> None:
    group = parser.add_argument_group("Learning rate and regularization parameters")
    group.add_argument("--lr", type=float, default=0.1, metavar="LR", help="base learning rate")
    group.add_argument("--bias-lr", type=float, metavar="LR", help="learning rate of biases")
    if backbone_lr is True:
        group.add_argument("--backbone-lr", type=float, metavar="LR", help="backbone learning rate")

    group.add_argument(
        "--lr-scale", type=int, help="reference batch size for LR scaling, if provided, LR will be scaled accordingly"
    )
    group.add_argument(
        "--lr-scale-type", type=str, choices=["linear", "sqrt"], default="linear", help="learning rate scaling type"
    )
    group.add_argument("--wd", type=float, default=0.0001, metavar="WD", help="weight decay")
    if wd_end is True:
        group.add_argument(
            "--wd-end", type=float, metavar="WD", help="final value of the weight decay (None for constant wd)"
        )

    group.add_argument("--norm-wd", type=float, metavar="WD", help="weight decay for Normalization layers")
    group.add_argument(
        "--bias-weight-decay", type=float, metavar="WD", help="weight decay for bias parameters of all layers"
    )
    group.add_argument(
        "--transformer-embedding-decay",
        type=float,
        metavar="WD",
        help="weight decay for embedding parameters for vision transformer models",
    )
    group.add_argument("--layer-decay", type=float, help="layer-wise learning rate decay (LLRD)")
    group.add_argument("--layer-decay-min-scale", type=float, help="minimum layer scale factor clamp value")
    group.add_argument(
        "--layer-decay-no-opt-scale", type=float, help="layer scale threshold below which parameters are frozen"
    )


def add_lr_scheduler_args(parser: argparse.ArgumentParser) -> None:
    group = parser.add_argument_group("Learning rate scheduler parameters")
    group.add_argument(
        "--lr-scheduler-update",
        type=str,
        choices=["epoch", "iter"],
        default="epoch",
        help="when to apply learning rate scheduler update",
    )
    group.add_argument(
        "--lr-scheduler",
        type=str,
        choices=list(get_args(SchedulerType)),
        default="constant",
        help="learning rate scheduler",
    )
    group.add_argument(
        "--lr-step-size",
        type=int,
        default=40,
        metavar="N",
        help="decrease lr every step-size epochs (for step scheduler only)",
    )
    group.add_argument(
        "--lr-steps", type=int, nargs="+", help="decrease lr every step-size epochs (multistep scheduler only)"
    )
    group.add_argument(
        "--lr-step-gamma",
        type=float,
        default=0.75,
        help="multiplicative factor of learning rate decay (for step scheduler only)",
    )
    group.add_argument(
        "--lr-cosine-min",
        type=float,
        default=0.000001,
        help="minimum learning rate (for cosine annealing scheduler only)",
    )
    group.add_argument(
        "--lr-power", type=float, default=1.0, help="power of the polynomial (for polynomial scheduler only)"
    )
    group.add_argument(
        "--lr-warmup-decay",
        type=float,
        default=0.01,
        help="multiplicative factor for learning rate at the start of warmup",
    )


def add_input_args(parser: argparse.ArgumentParser, size_help: Optional[str] = None) -> None:
    group = parser.add_argument_group("Input parameters")
    if size_help is None:
        size_help = "image size (defaults to the network default size)"

    group.add_argument("--channels", type=int, default=3, metavar="N", help="no. of image channels")
    group.add_argument("--size", type=int, nargs="+", metavar=("H", "W"), help=size_help)


def add_detection_input_args(parser: argparse.ArgumentParser) -> None:
    group = parser.add_argument_group("Input parameters")
    group.add_argument("--channels", type=int, default=3, metavar="N", help="no. of image channels")
    group.add_argument(
        "--size",
        type=int,
        nargs="+",
        metavar=("H", "W"),
        help=(
            "target image size as [height, width], if --dynamic-size is enabled, "
            "uses the smaller dimension as target size while preserving aspect ratio (defaults to model's signature)"
        ),
    )
    group.add_argument(
        "--max-size",
        type=int,
        help="maximum size for the longer edge of resized images, when specified, enables dynamic sizing",
    )
    group.add_argument(
        "--dynamic-size",
        default=False,
        action="store_true",
        help="allow variable image sizes while preserving aspect ratios",
    )
    group.add_argument("--multiscale", default=False, action="store_true", help="enable random scale per image")


def add_training_schedule_args(parser: argparse.ArgumentParser, default_epochs: int = 100) -> None:
    group = parser.add_argument_group("Training schedule parameters")
    group.add_argument("--epochs", type=int, default=default_epochs, metavar="N", help="number of training epochs")
    group.add_argument(
        "--stop-epoch", type=int, metavar="N", help="epoch to stop the training at (multi stage training)"
    )
    group.add_argument("--warmup-epochs", type=int, metavar="N", help="number of warmup epochs")
    group.add_argument("--warmup-iters", type=int, metavar="N", help="number of warmup iterations")
    group.add_argument("--cooldown-epochs", type=int, metavar="N", help="number of cooldown epochs (linear to zero)")
    group.add_argument("--cooldown-iters", type=int, metavar="N", help="number of cooldown iterations (linear to zero)")


def add_batch_norm_args(parser: argparse.ArgumentParser, backbone_freeze: bool = False) -> None:
    group = parser.add_argument_group("Batch normalization parameters")
    group.add_argument(
        "--freeze-bn",
        default=False,
        action="store_true",
        help="freeze all batch statistics and affine parameters of batchnorm2d layers",
    )
    if backbone_freeze is True:
        group.add_argument(
            "--freeze-backbone-bn",
            default=False,
            action="store_true",
            help="freeze all batch statistics and affine parameters of batchnorm2d layers (backbone only)",
        )

    group.add_argument("--sync-bn", default=False, action="store_true", help="use synchronized BatchNorm")


def add_data_aug_args(
    parser: argparse.ArgumentParser,
    default_level: int = 4,
    default_min_scale: Optional[float] = None,
    default_re_prob: Optional[float] = None,
    mixup_cutmix: bool = False,
) -> None:
    group = parser.add_argument_group("Data augmentation parameters")
    group.add_argument(
        "--aug-type", type=str, choices=list(get_args(AugType)), default="birder", help="augmentation type"
    )
    group.add_argument(
        "--aug-level",
        type=int,
        choices=list(range(10 + 1)),
        default=default_level,
        help="magnitude of birder augmentations (0 off -> 10 highest)",
    )
    group.add_argument(
        "--use-grayscale", default=False, action="store_true", help="use grayscale augmentation (birder aug only)"
    )
    group.add_argument(
        "--ra-num-ops",
        type=int,
        default=2,
        metavar="N",
        help="number of augmentation transformations to apply sequentially",
    )
    group.add_argument("--ra-magnitude", type=int, default=9, help="magnitude for all the RandAugment transformations")
    group.add_argument("--augmix-severity", type=int, default=3, help="severity of AugMix policy")
    group.add_argument("--resize-min-scale", type=float, default=default_min_scale, help="random resize min scale")
    group.add_argument(
        "--re-prob",
        type=float,
        default=default_re_prob,
        help="random erase probability (default according to aug-level)",
    )
    group.add_argument(
        "--simple-crop", default=False, action="store_true", help="use simple random crop (SRC) instead of RRC"
    )
    if mixup_cutmix is True:
        group.add_argument("--mixup-alpha", type=float, help="mixup alpha")
        group.add_argument("--cutmix", default=False, action="store_true", help="enable cutmix")

    group.add_argument(
        "--rgb-mode",
        type=str,
        choices=list(typing.get_args(RGBMode)),
        default="birder",
        help="RGB mean and std to use for normalization",
    )
    group.add_argument(
        "--rgb-mean",
        type=float,
        nargs=3,
        metavar=("R", "G", "B"),
        help="set custom RGB mean values (overrides values from selected RGB mode)",
    )
    group.add_argument(
        "--rgb-std",
        type=float,
        nargs=3,
        metavar=("R", "G", "B"),
        help="set custom RGB std values (overrides values from selected RGB mode)",
    )


def add_detection_data_aug_args(parser: argparse.ArgumentParser, default_level: int = 4) -> None:
    group = parser.add_argument_group("Data augmentation parameters")
    group.add_argument(
        "--aug-type", type=str, choices=list(get_args(DetAugType)), default="birder", help="augmentation type"
    )
    group.add_argument(
        "--aug-level",
        type=int,
        choices=list(range(10 + 1)),
        default=default_level,
        help="magnitude of birder augmentations (0 off -> 10 highest)",
    )
    group.add_argument(
        "--rgb-mode",
        type=str,
        choices=list(typing.get_args(RGBMode)),
        default="birder",
        help="RGB mean and std to use for normalization",
    )
    group.add_argument(
        "--rgb-mean",
        type=float,
        nargs=3,
        metavar=("R", "G", "B"),
        help="set custom RGB mean values (overrides values from selected RGB mode)",
    )
    group.add_argument(
        "--rgb-std",
        type=float,
        nargs=3,
        metavar=("R", "G", "B"),
        help="set custom RGB std values (overrides values from selected RGB mode)",
    )


def add_checkpoint_args(
    parser: argparse.ArgumentParser, default_save_frequency: int = 1, pretrained: bool = False
) -> None:
    group = parser.add_argument_group("Checkpoint parameters")
    group.add_argument(
        "--save-frequency", type=int, default=default_save_frequency, metavar="N", help="frequency of model saving"
    )
    group.add_argument(
        "--keep-last", type=int, metavar="N", help="number of recent checkpoints to keep (older ones are deleted)"
    )
    if pretrained is True:
        group.add_argument(
            "--pretrained",
            default=False,
            action="store_true",
            help="start with pretrained version of specified network (will download if not found locally)",
        )

    group.add_argument("--resume-epoch", type=int, metavar="N", help="epoch number to resume training from")
    group.add_argument(
        "--non-strict-weights",
        default=False,
        action="store_true",
        help="allow non-strict loading of model weights (missing or unexpected keys in state_dict)",
    )
    group.add_argument(
        "--load-states",
        default=False,
        action="store_true",
        help="load optimizer, scheduler and scaler states when resuming",
    )
    group.add_argument("--load-scheduler", default=False, action="store_true", help="load only scheduler when resuming")


def add_ema_args(
    parser: argparse.ArgumentParser, default_ema_steps: int = 32, default_ema_decay: float = 0.9999
) -> None:
    group = parser.add_argument_group("Exponential moving average parameters")
    group.add_argument(
        "--model-ema",
        default=False,
        action="store_true",
        help="enable tracking exponential moving average of model parameters",
    )
    group.add_argument(
        "--model-ema-steps",
        type=int,
        default=default_ema_steps,
        metavar="N",
        help="the number of iterations that controls how often to update the EMA model (adjusted to grad-accum-steps)",
    )
    group.add_argument(
        "--model-ema-decay",
        type=float,
        default=default_ema_decay,
        help="decay factor for exponential moving average of model parameters",
    )
    group.add_argument(
        "--model-ema-warmup",
        type=int,
        metavar="N",
        help="number of steps before EMA is applied (defaults to schedule warmup steps, pass 0 to disable warmup)",
    )


def add_dataloader_args(
    parser: argparse.ArgumentParser,
    no_img_loader: bool = False,
    default_num_workers: Optional[int] = None,
    default_drop_last: bool = False,
    ra_sampler: bool = False,
) -> None:
    group = parser.add_argument_group("Dataloader parameters")
    if no_img_loader is False:
        group.add_argument(
            "--img-loader", type=str, choices=["tv", "pil"], default="tv", help="backend to load and decode images"
        )

    if default_num_workers is None:
        default_num_workers = min(12, max(os.cpu_count() // 4, 4))  # type: ignore[operator]

    group.add_argument(
        "-j",
        "--num-workers",
        type=int,
        default=default_num_workers,
        metavar="N",
        help="number of preprocessing workers",
    )
    group.add_argument(
        "--prefetch-factor", type=int, metavar="N", help="number of batches loaded in advance by each worker"
    )
    if default_drop_last is True:
        group.add_argument(
            "--no-drop-last",
            dest="drop_last",
            default=True,
            action="store_false",
            help="drop the last incomplete batch",
        )
    else:
        group.add_argument("--drop-last", default=False, action="store_true", help="drop the last incomplete batch")

    if ra_sampler is True:
        group.add_argument(
            "--ra-sampler", default=False, action="store_true", help="whether to use Repeated Augmentation in training"
        )
        group.add_argument(
            "--ra-reps", type=int, default=3, metavar="N", help="number of repetitions for Repeated Augmentation"
        )


def add_precision_args(parser: argparse.ArgumentParser) -> None:
    group = parser.add_argument_group("Precision parameters")
    group.add_argument(
        "--model-dtype",
        type=str,
        choices=["float32", "float16", "bfloat16"],
        default="float32",
        help="model dtype to use",
    )
    group.add_argument(
        "--amp",
        default=False,
        action="store_true",
        help="enable automatic mixed precision (AMP) training via torch.amp",
    )
    group.add_argument(
        "--amp-dtype",
        type=str,
        choices=["float16", "bfloat16"],
        default="float16",
        help="whether to use float16 or bfloat16 for mixed precision",
    )
    group.add_argument(
        "--fast-matmul", default=False, action="store_true", help="use fast matrix multiplication (affects precision)"
    )


def add_distributed_args(parser: argparse.ArgumentParser) -> None:
    group = parser.add_argument_group("Distributed training parameters")
    group.add_argument("--world-size", type=int, default=1, metavar="N", help="number of distributed processes")
    group.add_argument("--local-rank", type=int, metavar="N", help="local rank")
    group.add_argument("--dist-url", type=str, default="env://", help="URL used to initialize distributed training")
    group.add_argument("--dist-backend", type=str, default="nccl", help="distributed backend")
    group.add_argument(
        "--find-unused-parameters",
        default=False,
        action="store_true",
        help="enable searching for unused parameters in DistributedDataParallel (may impact performance)",
    )


def add_logging_and_debug_args(
    parser: argparse.ArgumentParser, default_log_interval: int = 50, fake_data: bool = True
) -> None:
    group = parser.add_argument_group("Logging and debugging parameters")
    group.add_argument(
        "--experiment",
        "--exp",
        type=str,
        metavar="NAME",
        help="experiment name for logging (creates dedicated directory for the run)",
    )
    group.add_argument(
        "--log-interval",
        type=int,
        default=default_log_interval,
        metavar="N",
        help="how many steps between summary writes",
    )
    group.add_argument(
        "--grad-anomaly-detection",
        default=False,
        action="store_true",
        help="enable the autograd anomaly detection (for debugging)",
    )
    group.add_argument(
        "--use-deterministic-algorithms", default=False, action="store_true", help="use only deterministic algorithms"
    )
    group.add_argument(
        "--plot-lr", default=False, action="store_true", help="plot learning rate and exit (skip training)"
    )
    group.add_argument("--no-summary", default=False, action="store_true", help="don't print model summary")
    group.add_argument(
        "--non-interactive",
        default=False,
        action="store_true",
        help="force non-interactive mode (disables progress bars)",
    )
    group.add_argument(
        "--seed", type=int, help="set random seed for better reproducibility (affects torch, numpy and random)"
    )
    group.add_argument("--cpu", default=False, action="store_true", help="use cpu (mostly for testing)")
    if fake_data is True:
        group.add_argument(
            "--use-fake-data",
            default=False,
            action="store_true",
            help="use fake data instead of real dataset (like torchvision.datasets.FakeData)",
        )


def add_training_data_args(parser: argparse.ArgumentParser, unsupervised: bool = False) -> None:
    group = parser.add_argument_group("Training data parameters", description="WebDataset")
    group.add_argument("--wds", default=False, action="store_true", help="use webdataset for training")
    group.add_argument("--wds-info", type=str, metavar="FILE", help="wds info file path")
    group.add_argument("--wds-cache-dir", type=str, metavar="DIR", help="webdataset cache directory")
    group.add_argument("--wds-train-size", type=int, metavar="N", help="size of the wds training set")
    if unsupervised is False:
        group.add_argument("--wds-class-file", type=str, metavar="FILE", help="class list file")
        group.add_argument("--wds-val-size", type=int, metavar="N", help="size of the wds validation set")
        group.add_argument(
            "--wds-training-split", type=str, default="training", metavar="NAME", help="wds dataset train split"
        )
        group.add_argument(
            "--wds-val-split", type=str, default="validation", metavar="NAME", help="wds dataset validation split"
        )
    else:
        group.add_argument(
            "--wds-split", type=str, default="training", metavar="NAME", help="wds dataset split to load"
        )

    group.add_argument(
        "--wds-extra-shuffle",
        default=False,
        action="store_true",
        help=(
            "enable cross-worker batch shuffling after batching. Provides maximum sample diversity but incurs a "
            "notable performance penalty. Use with caution"
        ),
    )

    group = parser.add_argument_group(description="Directory")
    if unsupervised is False:
        group.add_argument(
            "--hierarchical",
            default=False,
            action="store_true",
            help="use hierarchical directory structure for labels (e.g., 'dir1/subdir2' -> 'dir1_subdir2' label)",
        )
        group.add_argument(
            "--data-path",
            type=str,
            default=str(settings.TRAINING_DATA_PATH),
            metavar="DIR",
            help="training directory path",
        )
        group.add_argument(
            "--val-path",
            type=str,
            default=str(settings.VALIDATION_DATA_PATH),
            metavar="DIR",
            help="validation directory path",
        )
    else:
        group.add_argument(
            "--data-path", nargs="*", default=[], help="training directories paths (directories and files)"
        )


def add_detection_training_data_args(parser: argparse.ArgumentParser) -> None:
    group = parser.add_argument_group("Training data parameters")
    group.add_argument(
        "--data-path",
        type=str,
        default=str(settings.DETECTION_DATA_PATH),
        metavar="DIR",
        help="training base directory path",
    )
    group.add_argument(
        "--val-path",
        type=str,
        default=str(settings.DETECTION_DATA_PATH),
        metavar="DIR",
        help="validation base directory path",
    )
    group.add_argument(
        "--coco-json-path",
        type=str,
        default=f"{settings.TRAINING_DETECTION_ANNOTATIONS_PATH}_coco.json",
        metavar="FILE",
        help="training COCO json path",
    )
    group.add_argument(
        "--coco-val-json-path",
        type=str,
        default=f"{settings.VALIDATION_DETECTION_ANNOTATIONS_PATH}_coco.json",
        metavar="FILE",
        help="validation COCO json path",
    )
    group.add_argument("--class-file", type=str, metavar="FILE", help="class list file, overrides json categories")
    group.add_argument(
        "--ignore-file", type=str, metavar="FILE", help="ignore list file, list of samples to ignore in training"
    )


# pylint: disable=too-many-branches
def common_args_validation(args: argparse.Namespace) -> None:
    # Some scripts like train_kd do not have a network argument,
    # but if it exists, it's mandatory all around
    if hasattr(args, "network") is True and args.network is None:
        raise ValidationError("--network is required")

    # Training schedule args, shared by all scripts
    if args.stop_epoch is not None and args.stop_epoch > args.epochs:
        raise ValidationError(
            f"--stop-epoch must be smaller than the total number of epochs ({args.epochs}), got {args.stop_epoch}"
        )
    if args.warmup_epochs is not None and args.warmup_iters is not None:
        raise ValidationError("--warmup-epochs cannot be used with --warmup-iters")
    if args.cooldown_epochs is not None and args.cooldown_iters is not None:
        raise ValidationError("--cooldown-epochs cannot be used with --cooldown-iters")

    if hasattr(args, "lr_scheduler_update") is True:
        if args.lr_scheduler_update != "iter" and args.warmup_iters is not None:
            raise ValidationError(
                "--warmup-iters can only be used when --lr-scheduler-update is 'iter', "
                f"but it is set to '{args.lr_scheduler_update}'"
            )
        if args.lr_scheduler_update != "iter" and args.cooldown_iters is not None:
            raise ValidationError(
                "--cooldown-iters can only be used when --lr-scheduler-update is 'iter', "
                f"but it is set to '{args.lr_scheduler_update}'"
            )

    # Compile args, argument dependant
    if hasattr(args, "compile_teacher") is True:
        if args.compile is True and args.compile_teacher is True:
            raise ValidationError("--compile cannot be used with --compile-teacher")
    if hasattr(args, "compile_backbone") is True:
        if args.compile is True and args.compile_backbone is True:
            raise ValidationError("--compile cannot be used with --compile-backbone")

    # Checkpoint args, shared by all scripts
    if args.load_states is True and args.resume_epoch is None:
        raise ValidationError("--load-states requires --resume-epoch to be set")
    if args.load_scheduler is True and args.resume_epoch is None:
        raise ValidationError("--load-scheduler requires --resume-epoch to be set")
    if hasattr(args, "pretrained") is True and args.pretrained is True and args.resume_epoch is not None:
        raise ValidationError("--pretrained cannot be used with --resume-epoch")

    # Data augmentation args have standard and detection version. Apply only to standard
    if hasattr(args, "resize_min_scale") is True:
        if args.resize_min_scale is not None and (args.resize_min_scale <= 0.0 or args.resize_min_scale >= 1.0):
            raise ValidationError(f"--resize-min-scale must be in range of (0, 1.0), got {args.resize_min_scale}")

    # Training data args have a standard and a detection version. The detection version does not have WDS
    if hasattr(args, "wds") is True:
        # Supervised WDS
        if hasattr(args, "wds_class_file") is True and args.wds is True and args.wds_class_file is None:
            raise ValidationError("--wds requires --wds-class-file to be set")

        if hasattr(args, "hierarchical") is True and args.wds is True and args.hierarchical is True:
            raise ValidationError("--wds cannot be used with --hierarchical")

        # WDS with debug args
        if hasattr(args, "use_fake_data") is True and args.use_fake_data is True and args.wds is True:
            raise ValidationError("--use-fake-data cannot be used with --wds")

        # WDS with dataloader args
        if hasattr(args, "ra_sampler") is True and args.wds is True and args.ra_sampler is True:
            raise ValidationError("Repeated Augmentation (--ra-sampler) not supported with WebDataset (--wds)")

        # Unsupervised training data
        if isinstance(args.data_path, list):
            if args.wds is False and len(args.data_path) == 0 and args.use_fake_data is False:
                raise ValidationError("Must provide at least one data source, --data-path or --wds")
            if args.wds is True and len(args.data_path) > 1:
                raise ValidationError(f"--wds can have at most 1 --data-path, got {len(args.data_path)}")

    # BatchNorm args, shared by some scripts
    if hasattr(args, "freeze_bn") is True and hasattr(args, "sync_bn"):
        if args.freeze_bn is True and args.sync_bn is True:
            raise ValidationError("--freeze-bn cannot be used with --sync-bn")

    # Precision_args, shared by all scripts
    if args.amp is True and args.model_dtype != "float32":
        raise ValidationError("--amp can only be used with --model-dtype float32")



import logging
import unittest

import torch

from birder.data.collators import detection

logging.disable(logging.CRITICAL)


class TestTransforms(unittest.TestCase):
    def test_detection(self) -> None:
        (images, masks, size_list) = detection.batch_images(
            [
                torch.ones((3, 10, 10)),
                torch.ones((3, 12, 12)),
            ],
            size_divisible=4,
        )

        self.assertSequenceEqual(images.size(), (2, 3, 12, 12))
        self.assertEqual(images[0][0][0][10].item(), 0)
        self.assertEqual(images[0][0][10][0].item(), 0)
        self.assertEqual(images[0][0][9][9].item(), 1)

        self.assertTrue(torch.all(masks[0, :10, :10] == False))  # pylint: disable=singleton-comparison # noqa: E712
        self.assertTrue(torch.all(masks[0, 11:, 11:] == True))  # pylint: disable=singleton-comparison # noqa: E712
        self.assertTrue(torch.all(masks[1] == False))  # pylint: disable=singleton-comparison # noqa: E712

        self.assertEqual(size_list[0], (10, 10))
        self.assertEqual(size_list[1], (12, 12))



"""
HGNet v2, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/hgnet.py

Reference document:
https://github.com/PaddlePaddle/PaddleClas/blob/develop/docs/zh_CN/models/ImageNet1k/PP-HGNetV2.md
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn

from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.hgnet_v1 import Conv2dBNActivation
from birder.net.hgnet_v1 import HighPerfGPUStage
from birder.net.hgnet_v1 import LearnableAffineBlock


class Stem(nn.Module):
    def __init__(self, in_channels: int, mid_channels: int, out_channels: int, use_lab: bool) -> None:
        super().__init__()
        self.stem1 = Conv2dBNActivation(
            in_channels,
            mid_channels,
            kernel_size=(3, 3),
            stride=(2, 2),
            padding=(1, 1),
            groups=1,
            use_act=True,
            use_lab=use_lab,
        )
        self.stem2a = Conv2dBNActivation(
            mid_channels,
            mid_channels // 2,
            kernel_size=(2, 2),
            stride=(1, 1),
            padding=(0, 0),
            groups=1,
            use_act=True,
            use_lab=use_lab,
        )
        self.stem2b = Conv2dBNActivation(
            mid_channels // 2,
            mid_channels,
            kernel_size=(2, 2),
            stride=(1, 1),
            padding=(0, 0),
            groups=1,
            use_act=True,
            use_lab=use_lab,
        )
        self.stem3 = Conv2dBNActivation(
            mid_channels * 2,
            mid_channels,
            kernel_size=(3, 3),
            stride=(2, 2),
            padding=(1, 1),
            groups=1,
            use_act=True,
            use_lab=use_lab,
        )
        self.stem4 = Conv2dBNActivation(
            mid_channels,
            out_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            groups=1,
            use_act=True,
            use_lab=use_lab,
        )
        self.pool = nn.MaxPool2d(kernel_size=(2, 2), stride=(1, 1), padding=(0, 0), ceil_mode=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem1(x)
        x = F.pad(x, (0, 1, 0, 1))
        x1 = self.pool(x)
        x2 = self.stem2a(x)
        x2 = F.pad(x2, (0, 1, 0, 1))
        x2 = self.stem2b(x2)

        x = torch.concat([x1, x2], dim=1)
        x = self.stem3(x)
        x = self.stem4(x)

        return x


# pylint: disable=invalid-name
class HGNet_v2(DetectorBackbone):
    # pylint: disable=too-many-locals
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        use_ese = False
        feature_dim = 2048
        stem_channels: list[int] = self.config["stem_channels"]
        stage1: list[int | bool] = self.config["stage1"]
        stage2: list[int | bool] = self.config["stage2"]
        stage3: list[int | bool] = self.config["stage3"]
        stage4: list[int | bool] = self.config["stage4"]
        use_lab: bool = self.config["use_lab"]
        drop_path_rate: float = self.config["drop_path_rate"]

        stages_cfg = [stage1, stage2, stage3, stage4]
        block_depths = [c[3] for c in stages_cfg]

        self.stem = Stem(self.input_channels, stem_channels[0], stem_channels[1], use_lab=use_lab)

        # Stochastic depth decay rule
        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(block_depths)).split(block_depths)]

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i, stage_config in enumerate(stages_cfg):
            (in_ch, mid_ch, out_ch, blocks, downsample, light_block, k, num_layers) = stage_config
            stages[f"stage{i+1}"] = HighPerfGPUStage(
                in_channels=in_ch,
                mid_channels=mid_ch,
                out_channels=out_ch,
                depth=blocks,
                num_layers=num_layers,
                downsample=downsample,  # type: ignore[arg-type]
                stride=(2, 2),
                kernel_size=(k, k),
                light_block=light_block,  # type: ignore[arg-type]
                use_lab=use_lab,
                use_ese=use_ese,
                drop_path=dpr[i],
            )
            return_channels.append(out_ch)

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Conv2d(out_ch, feature_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False),
            nn.ReLU(),
            LearnableAffineBlock(scale_value=1.0, bias_value=0.0) if use_lab is True else nn.Identity(),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = feature_dim
        self.classifier = self.create_classifier()

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "hgnet_v2_b0",
    HGNet_v2,
    config={
        "stem_channels": [16, 16],
        # in, mid, out, blocks, downsample, light_block, kernel_size, num_layers
        "stage1": [16, 16, 64, 1, False, False, 3, 3],
        "stage2": [64, 32, 256, 1, True, False, 3, 3],
        "stage3": [256, 64, 512, 2, True, True, 5, 3],
        "stage4": [512, 128, 1024, 1, True, True, 5, 3],
        "use_lab": True,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "hgnet_v2_b1",
    HGNet_v2,
    config={
        "stem_channels": [24, 32],
        # in, mid, out, blocks, downsample, light_block, kernel_size, num_layers
        "stage1": [32, 32, 64, 1, False, False, 3, 3],
        "stage2": [64, 48, 256, 1, True, False, 3, 3],
        "stage3": [256, 96, 512, 2, True, True, 5, 3],
        "stage4": [512, 192, 1024, 1, True, True, 5, 3],
        "use_lab": True,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "hgnet_v2_b2",
    HGNet_v2,
    config={
        "stem_channels": [24, 32],
        # in, mid, out, blocks, downsample, light_block, kernel_size, num_layers
        "stage1": [32, 32, 96, 1, False, False, 3, 4],
        "stage2": [96, 64, 384, 1, True, False, 3, 4],
        "stage3": [384, 128, 768, 3, True, True, 5, 4],
        "stage4": [768, 256, 1536, 1, True, True, 5, 4],
        "use_lab": True,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "hgnet_v2_b3",
    HGNet_v2,
    config={
        "stem_channels": [24, 32],
        # in, mid, out, blocks, downsample, light_block, kernel_size, num_layers
        "stage1": [32, 32, 128, 1, False, False, 3, 5],
        "stage2": [128, 64, 512, 1, True, False, 3, 5],
        "stage3": [512, 128, 1024, 3, True, True, 5, 5],
        "stage4": [1024, 256, 2048, 1, True, True, 5, 5],
        "use_lab": True,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "hgnet_v2_b4",
    HGNet_v2,
    config={
        "stem_channels": [32, 48],
        # in, mid, out, blocks, downsample, light_block, kernel_size, num_layers
        "stage1": [48, 48, 128, 1, False, False, 3, 6],
        "stage2": [128, 96, 512, 1, True, False, 3, 6],
        "stage3": [512, 192, 1024, 3, True, True, 5, 6],
        "stage4": [1024, 384, 2048, 1, True, True, 5, 6],
        "use_lab": False,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "hgnet_v2_b5",
    HGNet_v2,
    config={
        "stem_channels": [32, 64],
        # in, mid, out, blocks, downsample, light_block, kernel_size, num_layers
        "stage1": [64, 64, 128, 1, False, False, 3, 6],
        "stage2": [128, 128, 512, 2, True, False, 3, 6],
        "stage3": [512, 256, 1024, 5, True, True, 5, 6],
        "stage4": [1024, 512, 2048, 2, True, True, 5, 6],
        "use_lab": False,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "hgnet_v2_b6",
    HGNet_v2,
    config={
        "stem_channels": [48, 96],
        # in, mid, out, blocks, downsample, light_block, kernel_size, num_layers
        "stage1": [96, 96, 192, 2, False, False, 3, 6],
        "stage2": [192, 192, 512, 3, True, False, 3, 6],
        "stage3": [512, 384, 1024, 6, True, True, 5, 6],
        "stage4": [1024, 768, 2048, 3, True, True, 5, 6],
        "use_lab": False,
        "drop_path_rate": 0.0,
    },
)

registry.register_weights(
    "hgnet_v2_b0_il-common",
    {
        "description": "HGNet v2 B0 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 18.1,
                "sha256": "bb1a45bb4f7612ca60d524c5c31e42b445c75bfac4142b029f0523c7e6a697ff",
            }
        },
        "net": {"network": "hgnet_v2_b0", "tag": "il-common"},
    },
)
registry.register_weights(
    "hgnet_v2_b2_danube-delta",
    {
        "url": "https://huggingface.co/birder-project/hgnet_v2_b2_danube-delta/resolve/main",
        "description": "HGNet v2 B2 model trained on the danube-delta dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 38.1,
                "sha256": "9dc96728eb3085c21be34bbbb1b167eb836015e920b590e6dee40040ed2c887a",
            }
        },
        "net": {"network": "hgnet_v2_b2", "tag": "danube-delta"},
    },
)



import argparse
import json
import logging
import math
import os
import sys
import time
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.amp
import torchinfo
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from torchmetrics.detection.mean_ap import MeanAveragePrecision
from tqdm import tqdm

import birder
from birder.common import cli
from birder.common import fs_ops
from birder.common import lib
from birder.common import training_cli
from birder.common import training_utils
from birder.conf import settings
from birder.data.collators.detection import training_collate_fn
from birder.data.datasets.coco import CocoTraining
from birder.data.transforms.classification import get_rgb_stats
from birder.data.transforms.detection import InferenceTransform
from birder.data.transforms.detection import training_preset
from birder.model_registry import Task
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.detection.base import get_detection_signature

logger = logging.getLogger(__name__)


# pylint: disable=too-many-locals,too-many-branches,too-many-statements
def train(args: argparse.Namespace) -> None:
    #
    # Initialize
    #
    training_utils.init_distributed_mode(args)
    logger.info(f"Starting training, birder version: {birder.__version__}, pytorch version: {torch.__version__}")
    training_utils.log_git_info()

    if (
        args.multiscale is True
        or args.dynamic_size is True
        or args.max_size is not None
        or args.aug_type == "multiscale"
        or args.aug_type == "detr"
    ):
        dynamic_size = True
    else:
        dynamic_size = False

    if args.size is None:
        args.size = registry.get_default_size(args.network)

    if dynamic_size is False:
        logger.info(f"Using size={args.size}")
    else:
        logger.info(f"Running with dynamic size, with base size={args.size}")

    if args.cpu is True:
        device = torch.device("cpu")
        device_id = 0
    else:
        device = torch.device("cuda")
        device_id = torch.cuda.current_device()

    if args.use_deterministic_algorithms is True:
        torch.backends.cudnn.benchmark = False
        torch.use_deterministic_algorithms(True)
    elif dynamic_size is False:
        torch.backends.cudnn.benchmark = True

    if args.seed is not None:
        lib.set_random_seeds(args.seed)

    if args.non_interactive is True or training_utils.is_local_primary(args) is False:
        disable_tqdm = True
    elif sys.stderr.isatty() is False:
        disable_tqdm = True
    else:
        disable_tqdm = False

    # Enable or disable the autograd anomaly detection
    torch.autograd.set_detect_anomaly(args.grad_anomaly_detection)

    #
    # Data
    #
    rgb_stats = get_rgb_stats(args.rgb_mode, args.rgb_mean, args.rgb_std)
    logger.debug(f"Using RGB stats: {rgb_stats}")

    training_dataset = CocoTraining(
        args.data_path,
        args.coco_json_path,
        transforms=training_preset(
            args.size, args.aug_type, args.aug_level, rgb_stats, args.dynamic_size, args.multiscale, args.max_size
        ),
    )
    validation_dataset = CocoTraining(
        args.val_path,
        args.coco_val_json_path,
        transforms=InferenceTransform(args.size, rgb_stats, dynamic_size, args.max_size),
    )

    if args.class_file is not None:
        class_to_idx = fs_ops.read_class_file(args.class_file)
    else:
        class_to_idx = lib.class_to_idx_from_coco(training_dataset.dataset.coco.cats)

    class_to_idx = lib.detection_class_to_idx(class_to_idx)
    if args.ignore_file is not None:
        with open(args.ignore_file, "r", encoding="utf-8") as handle:
            ignore_list = handle.read().splitlines()
    else:
        ignore_list = []

    if args.binary_mode is True:
        training_dataset.convert_to_binary_annotations()
        validation_dataset.convert_to_binary_annotations()
        class_to_idx = training_dataset.class_to_idx

    training_dataset.remove_images_without_annotations(ignore_list)

    assert args.model_ema is False or args.model_ema_steps <= len(training_dataset) / args.batch_size

    logger.info(f"Using device {device}:{device_id}")
    logger.info(f"Training on {len(training_dataset):,} samples")
    logger.info(f"Validating on {len(validation_dataset):,} samples")

    num_outputs = len(class_to_idx)  # Does not include background class
    batch_size: int = args.batch_size
    model_ema_steps: int = args.model_ema_steps * args.grad_accum_steps
    logger.debug(f"Effective batch size = {args.batch_size * args.grad_accum_steps * args.world_size}")

    # Data loaders and samplers
    (train_sampler, validation_sampler) = training_utils.get_samplers(args, training_dataset, validation_dataset)

    training_loader = DataLoader(
        training_dataset,
        batch_size=batch_size,
        sampler=train_sampler,
        num_workers=args.num_workers,
        prefetch_factor=args.prefetch_factor,
        collate_fn=training_collate_fn,
        pin_memory=True,
        drop_last=args.drop_last,
    )
    validation_loader = DataLoader(
        validation_dataset,
        batch_size=batch_size,
        sampler=validation_sampler,
        num_workers=args.num_workers,
        prefetch_factor=args.prefetch_factor,
        collate_fn=training_collate_fn,
        pin_memory=True,
        drop_last=args.drop_last,
    )

    last_batch_idx = len(training_loader) - 1
    begin_epoch = 1
    epochs = args.epochs + 1
    if args.stop_epoch is None:
        args.stop_epoch = epochs
    else:
        args.stop_epoch += 1

    #
    # Initialize network
    #
    model_dtype: torch.dtype = getattr(torch, args.model_dtype)
    sample_shape = (batch_size, args.channels, *args.size)  # B, C, H, W
    network_name = lib.get_detection_network_name(
        args.network, tag=args.tag, backbone=args.backbone, backbone_tag=args.backbone_tag
    )

    if args.resume_epoch is not None:
        begin_epoch = args.resume_epoch + 1
        (net, class_to_idx_saved, training_states) = fs_ops.load_detection_checkpoint(
            device,
            args.network,
            config=args.model_config,
            tag=args.tag,
            backbone=args.backbone,
            backbone_config=args.backbone_model_config,
            backbone_tag=args.backbone_tag,
            epoch=args.resume_epoch,
            new_size=args.size,
            strict=not args.non_strict_weights,
        )
        if args.reset_head is True:
            net.reset_classifier(len(class_to_idx))
        else:
            assert class_to_idx == class_to_idx_saved

    elif args.pretrained is True:
        fs_ops.download_model_by_weights(network_name, progress_bar=training_utils.is_local_primary(args))
        (net, class_to_idx_saved, training_states) = fs_ops.load_detection_checkpoint(
            device,
            args.network,
            config=args.model_config,
            tag=args.tag,
            backbone=args.backbone,
            backbone_config=args.backbone_model_config,
            backbone_tag=args.backbone_tag,
            epoch=None,
            new_size=args.size,
            strict=not args.non_strict_weights,
        )
        if args.reset_head is True:
            net.reset_classifier(len(class_to_idx))
        else:
            assert class_to_idx == class_to_idx_saved

    else:
        if args.backbone_epoch is not None:
            backbone: DetectorBackbone
            (backbone, class_to_idx_saved, _) = fs_ops.load_checkpoint(
                device,
                args.backbone,
                config=args.backbone_model_config,
                tag=args.backbone_tag,
                epoch=args.backbone_epoch,
                new_size=args.size,
                strict=not args.non_strict_weights,
            )

        elif args.backbone_pretrained is True:
            fs_ops.download_model_by_weights(
                lib.get_network_name(args.backbone, tag=args.backbone_tag),
                progress_bar=training_utils.is_local_primary(args),
            )
            (backbone, class_to_idx_saved, _) = fs_ops.load_checkpoint(
                device,
                args.backbone,
                config=args.backbone_model_config,
                tag=args.backbone_tag,
                epoch=None,
                new_size=args.size,
                strict=not args.non_strict_weights,
            )

        else:
            backbone = registry.net_factory(
                args.backbone, sample_shape[1], num_outputs, config=args.backbone_model_config, size=args.size
            )

        net = registry.detection_net_factory(
            args.network, num_outputs, backbone, config=args.model_config, size=args.size
        )
        training_states = fs_ops.TrainingStates.empty()

    net.to(device, dtype=model_dtype)
    if dynamic_size is True:
        net.set_dynamic_size()

    # Freeze
    if args.freeze_body is True:
        net.freeze(freeze_classifier=False)
    elif args.freeze_backbone is True:
        net.backbone.freeze()
    elif args.freeze_backbone_stages is not None:
        net.backbone.freeze_stages(up_to_stage=args.freeze_backbone_stages)

    if args.freeze_bn is True:
        net = training_utils.freeze_batchnorm2d(net)
    elif args.freeze_backbone_bn is True:
        net.backbone = training_utils.freeze_batchnorm2d(net.backbone)

    if args.sync_bn is True and args.distributed is True:
        net = torch.nn.SyncBatchNorm.convert_sync_batchnorm(net)

    if args.fast_matmul is True or args.amp is True:
        torch.set_float32_matmul_precision("high")

    # Compile network
    if args.compile is True:
        net = torch.compile(net)
    elif args.compile_backbone is True:
        net.backbone.detection_features = torch.compile(net.backbone.detection_features)  # type: ignore[method-assign]

    #
    # Loss criteria, optimizer, learning rate scheduler and training parameter groups
    #

    # Training parameter groups
    custom_keys_weight_decay = training_utils.get_wd_custom_keys(args)
    parameters = training_utils.optimizer_parameter_groups(
        net,
        args.wd,
        norm_weight_decay=args.norm_wd,
        custom_keys_weight_decay=custom_keys_weight_decay,
        layer_decay=args.layer_decay,
        layer_decay_min_scale=args.layer_decay_min_scale,
        layer_decay_no_opt_scale=args.layer_decay_no_opt_scale,
        bias_lr=args.bias_lr,
        backbone_lr=args.backbone_lr,
    )

    # Learning rate scaling
    lr = training_utils.scale_lr(args)
    grad_accum_steps: int = args.grad_accum_steps

    if args.lr_scheduler_update == "epoch":
        iter_update = False
        iters_per_epoch = 1
    elif args.lr_scheduler_update == "iter":
        iter_update = True
        iters_per_epoch = math.ceil(len(training_loader) / grad_accum_steps)
    else:
        raise ValueError("Unsupported lr_scheduler_update")

    # Optimizer and learning rate scheduler
    optimizer = training_utils.get_optimizer(parameters, lr, args)
    scheduler = training_utils.get_scheduler(optimizer, iters_per_epoch, args)
    if args.compile_opt is True:
        optimizer.step = torch.compile(optimizer.step, fullgraph=False)

    # Gradient scaler and AMP related tasks
    (scaler, amp_dtype) = training_utils.get_amp_scaler(args.amp, args.amp_dtype)

    # Load states
    if args.load_states is True:
        optimizer.load_state_dict(training_states.optimizer_state)
        scheduler.load_state_dict(training_states.scheduler_state)
        if scaler is not None:
            scaler.load_state_dict(training_states.scaler_state)

    elif args.load_scheduler is True:
        scheduler.load_state_dict(training_states.scheduler_state)
        last_lrs = scheduler.get_last_lr()
        for g, last_lr in zip(optimizer.param_groups, last_lrs):
            g["lr"] = last_lr

    last_lr = max(scheduler.get_last_lr())
    if args.plot_lr is True:
        logger.info("Fast forwarding scheduler...")
        optimizer.step()
        lrs = []
        for _ in range(begin_epoch, epochs):
            for _ in range(iters_per_epoch):
                lrs.append(max(scheduler.get_last_lr()))
                scheduler.step()

        plt.plot(np.linspace(begin_epoch, epochs, iters_per_epoch * (epochs - begin_epoch), endpoint=False), lrs)
        plt.show()
        raise SystemExit(0)

    #
    # Distributed (DDP) and Model EMA
    #
    if args.model_ema_warmup is not None:
        ema_warmup_epochs = args.model_ema_warmup
    elif args.warmup_epochs is not None:
        ema_warmup_epochs = args.warmup_epochs
    elif args.warmup_iters is not None:
        ema_warmup_epochs = args.warmup_iters // iters_per_epoch
    else:
        ema_warmup_epochs = 0

    logger.debug(f"EMA warmup epochs = {ema_warmup_epochs}")
    net_without_ddp = net
    if args.distributed is True:
        net = torch.nn.parallel.DistributedDataParallel(
            net, device_ids=[args.local_rank], find_unused_parameters=args.find_unused_parameters
        )
        net_without_ddp = net.module

    if args.model_ema is True:
        model_base = net_without_ddp  # Original model without DDP wrapper, will be saved as training state
        model_ema = training_utils.ema_model(args, net_without_ddp, device=device)
        if args.load_states is True and training_states.ema_model_state is not None:
            logger.info("Setting model EMA weights...")
            if args.compile is True and hasattr(model_ema.module, "_orig_mod") is True:
                model_ema.module._orig_mod.load_state_dict(  # pylint: disable=protected-access
                    training_states.ema_model_state
                )
            else:
                model_ema.module.load_state_dict(training_states.ema_model_state)

            model_ema.n_averaged += 1  # pylint:disable=no-member

        model_to_save = model_ema.module  # Save EMA model weights as default weights
        eval_model = model_ema  # Use EMA for evaluation

    else:
        model_base = None
        model_to_save = net_without_ddp
        eval_model = net

    if args.compile is True and hasattr(model_to_save, "_orig_mod") is True:
        model_to_save = model_to_save._orig_mod  # pylint: disable=protected-access
    if args.compile is True and hasattr(model_base, "_orig_mod") is True:
        model_base = model_base._orig_mod  # type: ignore[union-attr] # pylint: disable=protected-access

    #
    # Misc
    #

    # Define metrics
    validation_metrics = MeanAveragePrecision(iou_type="bbox", box_format="xyxy", average="macro").to(device)
    metric_list = ["map", "map_small", "map_medium", "map_large", "map_50", "map_75", "mar_1", "mar_10"]

    # Print network summary
    net_for_info = net_without_ddp
    if args.compile is True and hasattr(net_without_ddp, "_orig_mod") is True:
        net_for_info = net_without_ddp._orig_mod  # pylint: disable=protected-access

    if args.no_summary is False:
        summary = torchinfo.summary(
            net_for_info,
            device=device,
            input_size=sample_shape,
            dtypes=[model_dtype],
            col_names=["input_size", "output_size", "kernel_size", "num_params"],
            depth=4,
            verbose=0,
        )
        if training_utils.is_global_primary(args) is True:
            # Write to stderr, same as all the logs
            print(summary, file=sys.stderr)

    # Training logs
    training_log_path = training_utils.training_log_path(network_name, device, args.experiment)
    logger.info(f"Logging training run at {training_log_path}")
    summary_writer = SummaryWriter(training_log_path)

    signature = get_detection_signature(input_shape=sample_shape, num_outputs=num_outputs, dynamic=dynamic_size)
    file_handler: logging.Handler = logging.NullHandler()
    if training_utils.is_local_primary(args) is True:
        summary_writer.flush()
        fs_ops.write_config(network_name, net_for_info, signature=signature, rgb_stats=rgb_stats)
        file_handler = training_utils.setup_file_logging(training_log_path.joinpath("training.log"))
        with open(training_log_path.joinpath("training_args.json"), "w", encoding="utf-8") as handle:
            json.dump(
                {
                    "birder_version": birder.__version__,
                    "pytorch_version": torch.__version__,
                    "cmdline": " ".join(sys.argv),
                    **vars(args),
                },
                handle,
                indent=2,
            )

        with open(training_log_path.joinpath("training_data.json"), "w", encoding="utf-8") as handle:
            json.dump(
                {
                    "training_samples": len(training_dataset),
                    "validation_samples": len(validation_dataset),
                    "classes": list(class_to_idx.keys()),
                },
                handle,
                indent=2,
            )

    #
    # Training loop
    #
    logger.info(f"Starting training with learning rate of {last_lr}")
    for epoch in range(begin_epoch, args.stop_epoch):
        tic = time.time()
        net.train()
        running_loss = training_utils.SmoothedValue()
        validation_metrics.reset()

        if args.distributed is True:
            train_sampler.set_epoch(epoch)

        progress = tqdm(
            desc=f"Epoch {epoch}/{epochs-1}",
            total=len(training_dataset),
            leave=False,
            disable=disable_tqdm,
            unit="samples",
            initial=0,
        )

        # Zero the parameter gradients
        optimizer.zero_grad()

        epoch_start = time.time()
        start_time = epoch_start
        last_idx = 0
        for i, (inputs, targets, masks, image_sizes) in enumerate(training_loader):
            inputs = inputs.to(device, dtype=model_dtype, non_blocking=True)
            targets = [
                {k: v.to(device, non_blocking=True) if isinstance(v, torch.Tensor) else v for k, v in t.items()}
                for t in targets
            ]
            masks = masks.to(device, non_blocking=True)

            optimizer_update = (i == last_batch_idx) or ((i + 1) % grad_accum_steps == 0)

            # Forward, backward and optimize
            with torch.amp.autocast("cuda", enabled=args.amp, dtype=amp_dtype):
                (_detections, losses) = net(inputs, targets, masks, image_sizes)
                loss = sum(v for v in losses.values())

            if scaler is not None:
                scaler.scale(loss).backward()
                if optimizer_update is True:
                    if args.clip_grad_norm is not None:
                        scaler.unscale_(optimizer)
                        torch.nn.utils.clip_grad_norm_(net.parameters(), args.clip_grad_norm)

                    scaler.step(optimizer)
                    scaler.update()
                    optimizer.zero_grad()
                    if iter_update is True:
                        scheduler.step()

            else:
                loss.backward()
                if optimizer_update is True:
                    if args.clip_grad_norm is not None:
                        torch.nn.utils.clip_grad_norm_(net.parameters(), args.clip_grad_norm)

                    optimizer.step()
                    optimizer.zero_grad()
                    if iter_update is True:
                        scheduler.step()

            # Exponential moving average
            if args.model_ema is True and i % model_ema_steps == 0:
                model_ema.update_parameters(net)
                if epoch <= ema_warmup_epochs:
                    # Reset ema buffer to keep copying weights during warmup period
                    model_ema.n_averaged.fill_(0)  # pylint: disable=no-member

            # Statistics
            running_loss.update(loss.detach())

            # Write statistics
            if (i == last_batch_idx) or (i + 1) % args.log_interval == 0:
                time_now = time.time()
                time_cost = time_now - start_time
                steps_processed_in_interval = i - last_idx
                rate = steps_processed_in_interval * (batch_size * args.world_size) / time_cost

                avg_time_per_step = time_cost / steps_processed_in_interval
                remaining_steps_in_epoch = last_batch_idx - i
                estimated_time_to_finish_epoch = remaining_steps_in_epoch * avg_time_per_step

                start_time = time_now
                last_idx = i
                cur_lr = max(scheduler.get_last_lr())

                running_loss.synchronize_between_processes(device)
                with training_utils.single_handler_logging(logger, file_handler, enabled=not disable_tqdm) as log:
                    log.info(
                        f"[Trn] Epoch {epoch}/{epochs-1}, step {i+1}/{last_batch_idx+1}  "
                        f"Loss: {running_loss.avg:.4f}  "
                        f"Elapsed: {lib.format_duration(time_now-epoch_start)}  "
                        f"ETA: {lib.format_duration(estimated_time_to_finish_epoch)}  "
                        f"T: {time_cost:.1f}s  "
                        f"R: {rate:.1f} samples/s  "
                        f"LR: {cur_lr:.4e}"
                    )

                if training_utils.is_local_primary(args) is True:
                    summary_writer.add_scalars(
                        "loss",
                        {"training": running_loss.avg},
                        ((epoch - 1) * len(training_dataset)) + (i * batch_size * args.world_size),
                    )

            # Update progress bar
            progress.update(n=batch_size * args.world_size)

        progress.close()

        # Epoch training metrics
        logger.info(f"[Trn] Epoch {epoch}/{epochs-1} training_loss: {running_loss.global_avg:.4f}")

        # Validation
        eval_model.eval()
        progress = tqdm(
            desc=f"Epoch {epoch}/{epochs-1}",
            total=len(validation_dataset),
            leave=False,
            disable=disable_tqdm,
            unit="samples",
            initial=0,
        )
        with training_utils.single_handler_logging(logger, file_handler, enabled=not disable_tqdm) as log:
            log.info(f"[Val] Starting validation for epoch {epoch}/{epochs-1}...")

        epoch_start = time.time()
        with torch.inference_mode():
            for inputs, targets, masks, image_sizes in validation_loader:
                inputs = inputs.to(device, non_blocking=True)
                targets = [
                    {k: v.to(device, non_blocking=True) if isinstance(v, torch.Tensor) else v for k, v in t.items()}
                    for t in targets
                ]
                masks = masks.to(device, non_blocking=True)

                with torch.amp.autocast("cuda", enabled=args.amp, dtype=amp_dtype):
                    (detections, losses) = eval_model(inputs, masks=masks, image_sizes=image_sizes)

                for target in targets:
                    # TorchMetrics can't handle "empty" images
                    if "boxes" not in target:
                        target["boxes"] = torch.tensor([], dtype=torch.float, device=device)
                        target["labels"] = torch.tensor([], dtype=torch.int64, device=device)

                # Statistics
                validation_metrics(detections, targets)

                # Update progress bar
                if training_utils.is_local_primary(args) is True:
                    progress.update(n=batch_size * args.world_size)

        time_now = time.time()
        rate = len(validation_dataset) / (time_now - epoch_start)
        with training_utils.single_handler_logging(logger, file_handler, enabled=not disable_tqdm) as log:
            log.info(
                f"[Val] Epoch {epoch}/{epochs-1} "
                f"Elapsed: {lib.format_duration(time_now-epoch_start)}  "
                f"R: {rate:.1f} samples/s"
            )

        progress.close()

        validation_metrics_dict = validation_metrics.compute()

        # Write statistics
        if training_utils.is_local_primary(args) is True:
            for metric in metric_list:
                summary_writer.add_scalars(
                    "performance", {metric: validation_metrics_dict[metric]}, epoch * len(training_dataset)
                )

        # Epoch validation metrics
        for metric in metric_list:
            logger.info(f"[Val] Epoch {epoch}/{epochs-1} {metric}: {validation_metrics_dict[metric]:.4f}")

        # Learning rate scheduler update
        if iter_update is False:
            scheduler.step()
        if last_lr != max(scheduler.get_last_lr()):
            last_lr = max(scheduler.get_last_lr())
            logger.info(f"Updated learning rate to: {last_lr}")

        if training_utils.is_local_primary(args) is True:
            # Checkpoint model
            if epoch % args.save_frequency == 0:
                fs_ops.checkpoint_model(
                    network_name,
                    epoch,
                    model_to_save,
                    signature,
                    class_to_idx,
                    rgb_stats,
                    optimizer,
                    scheduler,
                    scaler,
                    model_base,
                )
                if args.keep_last is not None:
                    fs_ops.clean_checkpoints(network_name, args.keep_last)

        # Epoch timing
        toc = time.time()
        logger.info(f"Total time: {lib.format_duration(toc - tic)}")
        logger.info("---")

        # Reset counters
        epoch_start = time.time()
        start_time = epoch_start
        last_idx = 0

    # Save model hyperparameters with metrics
    if training_utils.is_local_primary(args) is True:
        # Replace list based args
        if args.opt_betas is not None:
            for idx, beta in enumerate(args.opt_betas):
                setattr(args, f"opt_betas_{idx}", beta)

            del args.opt_betas

        if args.lr_steps is not None:
            args.lr_steps = json.dumps(args.lr_steps)
        if args.model_config is not None:
            args.model_config = json.dumps(args.model_config)
        if args.backbone_model_config is not None:
            args.backbone_model_config = json.dumps(args.backbone_model_config)
        if args.size is not None:
            args.size = json.dumps(args.size)

        # Save all args
        val_metrics = validation_metrics.compute()
        summary_writer.add_hparams(
            {**vars(args), "training_samples": len(training_dataset)},
            {"hparam/val_map": val_metrics["map"]},
        )

    summary_writer.close()

    # Checkpoint model
    if training_utils.is_local_primary(args) is True:
        fs_ops.checkpoint_model(
            network_name,
            epoch,
            model_to_save,
            signature,
            class_to_idx,
            rgb_stats,
            optimizer,
            scheduler,
            scaler,
            model_base,
        )

    training_utils.shutdown_distributed_mode(args)


def get_args_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="Train object detection model",
        epilog=(
            "Usage examples\n"
            "==============\n"
            "A classic Faster-RCNN with EfficientNet v2 backbone:\n"
            "torchrun --nproc_per_node=2 -m birder.scripts.train_detection \\\n"
            "    --network faster_rcnn  \\\n"
            "    --backbone efficientnet_v2_s  \\\n"
            "    --backbone-epoch 0  \\\n"
            "    --lr 0.02  \\\n"
            "    --lr-scheduler multistep  \\\n"
            "    --lr-steps 16 22  \\\n"
            "    --lr-step-gamma 0.1  \\\n"
            "    --freeze-backbone-bn  \\\n"
            "    --batch-size 16  \\\n"
            "    --epochs 26  \\\n"
            "    --wd 0.0001  \\\n"
            "    --fast-matmul  \\\n"
            "    --compile\n"
            "\n"
            "A more modern Deformable-DETR example:\n"
            "torchrun --nproc_per_node=2 train_detection.py \\\n"
            "    --network deformable_detr \\\n"
            "    --backbone regnet_y_4g \\\n"
            "    --backbone-epoch 0 \\\n"
            "    --opt adamw \\\n"
            "    --lr 0.0002 \\\n"
            "    --backbone-lr 0.00002 \\\n"
            "    --lr-scheduler cosine \\\n"
            "    --freeze-backbone-bn \\\n"
            "    --batch-size 8 \\\n"
            "    --epochs 50 \\\n"
            "    --wd 0.0001 \\\n"
            "    --clip-grad-norm 1 \\\n"
            "    --fast-matmul \\\n"
            "    --compile-backbone \\\n"
            "    --compile-opt\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    parser.add_argument("-n", "--network", type=str, help="the neural network to use")
    parser.add_argument(
        "--model-config",
        action=cli.FlexibleDictAction,
        help=(
            "override the model default configuration, accepts key-value pairs or JSON "
            "('drop_path_rate=0.2' or '{\"units\": [3, 24, 36, 3], \"dropout\": 0.2}'"
        ),
    )
    parser.add_argument("-t", "--tag", type=str, help="add model tag")
    parser.add_argument("--backbone", type=str, help="the neural network to used as backbone")
    parser.add_argument(
        "--backbone-model-config",
        action=cli.FlexibleDictAction,
        help=(
            "override the backbone default configuration, accepts key-value pairs or JSON "
            "('drop_path_rate=0.2' or '{\"units\": [3, 24, 36, 3], \"dropout\": 0.2}'"
        ),
    )
    parser.add_argument("--backbone-tag", type=str, help="backbone training log tag (loading only)")
    parser.add_argument("--backbone-epoch", type=int, help="load backbone weights from selected epoch")
    parser.add_argument(
        "--backbone-pretrained",
        default=False,
        action="store_true",
        help="start with pretrained version of specified backbone (will download if not found locally)",
    )
    parser.add_argument("--reset-head", default=False, action="store_true", help="reset the classification head")
    parser.add_argument(
        "--freeze-body",
        default=False,
        action="store_true",
        help="freeze all layers of the model except the classification head",
    )
    parser.add_argument("--freeze-backbone", default=False, action="store_true", help="freeze backbone")
    parser.add_argument("--freeze-backbone-stages", type=int, help="number of backbone stages to freeze")
    parser.add_argument(
        "--binary-mode",
        default=False,
        action="store_true",
        help="treat all objects as a single class (binary detection: object vs background)",
    )
    training_cli.add_optimization_args(parser, default_batch_size=16)
    training_cli.add_lr_wd_args(parser, backbone_lr=True)
    training_cli.add_lr_scheduler_args(parser)
    training_cli.add_training_schedule_args(parser)
    training_cli.add_detection_input_args(parser)
    training_cli.add_detection_data_aug_args(parser)
    training_cli.add_ema_args(parser, default_ema_steps=1, default_ema_decay=0.9998)
    training_cli.add_dataloader_args(
        parser,
        no_img_loader=True,
        default_num_workers=min(8, max(os.cpu_count() // 8, 2)),  # type: ignore[operator]
        ra_sampler=True,
    )
    training_cli.add_batch_norm_args(parser, backbone_freeze=True)
    training_cli.add_precision_args(parser)
    training_cli.add_compile_args(parser, backbone=True)
    training_cli.add_checkpoint_args(parser, pretrained=True)
    training_cli.add_distributed_args(parser)
    training_cli.add_logging_and_debug_args(parser, default_log_interval=20, fake_data=False)
    training_cli.add_detection_training_data_args(parser)

    return parser


def validate_args(args: argparse.Namespace) -> None:
    args.data_path = str(args.data_path)
    args.val_path = str(args.val_path)
    args.size = cli.parse_size(args.size)

    # This will capture the common argument mistakes
    training_cli.common_args_validation(args)

    # Script specific checks
    if args.backbone is None:
        raise cli.ValidationError("--backbone is required")
    if registry.exists(args.network, task=Task.OBJECT_DETECTION) is False:
        raise cli.ValidationError(f"--network {args.network} not supported, see list-models tool for available options")
    if registry.exists(args.backbone, net_type=DetectorBackbone) is False:
        raise cli.ValidationError(
            f"--backbone {args.network} not supported, see list-models tool for available options"
        )

    if args.freeze_backbone is True and args.freeze_backbone_stages is not None:
        raise cli.ValidationError("--freeze-backbone cannot be used with --freeze-backbone-stages")
    if args.freeze_backbone is True and args.freeze_body is True:
        raise cli.ValidationError("--freeze-backbone cannot be used with --freeze-body")
    if args.freeze_body is True and args.freeze_backbone_stages is not None:
        raise cli.ValidationError("--freeze-body cannot be used with --freeze-backbone-stages")
    if args.multiscale is True and args.aug_type != "birder":
        raise cli.ValidationError(f"--multiscale only supported with --aug-type birder, got {args.aug_type}")
    if args.backbone_pretrained is True and args.backbone_epoch is not None:
        raise cli.ValidationError("--backbone-pretrained cannot be used with --backbone-epoch")

    if args.model_dtype != "float32":  # NOTE: only float32 supported at this time
        raise cli.ValidationError(f"Only float32 supported for --model-dtype at this time, got {args.model_dtype}")


def args_from_dict(**kwargs: Any) -> argparse.Namespace:
    parser = get_args_parser()
    parser.set_defaults(**kwargs)
    args = parser.parse_args([])
    validate_args(args)

    return args


def main() -> None:
    parser = get_args_parser()
    args = parser.parse_args()
    validate_args(args)

    if settings.MODELS_DIR.exists() is False:
        logger.info(f"Creating {settings.MODELS_DIR} directory...")
        settings.MODELS_DIR.mkdir(parents=True)

    train(args)


if __name__ == "__main__":
    logger = logging.getLogger(getattr(__spec__, "name", __name__))
    main()



"""
ShuffleNet v2, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/shufflenetv2.py

Paper "ShuffleNet V2: Practical Guidelines for Efficient CNN Architecture Design",
https://arxiv.org/abs/1807.11164
"""

# Reference license: BSD 3-Clause

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.shufflenet_v1 import channel_shuffle


class ShuffleUnit(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, dw_conv_stride: int) -> None:
        super().__init__()
        self.dw_conv_stride = dw_conv_stride
        branch_channels = out_channels // 2

        if dw_conv_stride == 1:
            branch2_input = branch_channels
            self.branch1 = nn.Sequential()

        else:
            branch2_input = in_channels
            self.branch1 = nn.Sequential(
                nn.Conv2d(
                    in_channels,
                    in_channels,
                    kernel_size=(3, 3),
                    stride=(dw_conv_stride, dw_conv_stride),
                    padding=(1, 1),
                    groups=in_channels,
                    bias=False,
                ),
                nn.BatchNorm2d(in_channels),
                Conv2dNormActivation(
                    in_channels,
                    branch_channels,
                    kernel_size=(1, 1),
                    stride=(1, 1),
                    padding=(0, 0),
                    bias=False,
                ),
            )

        self.branch2 = nn.Sequential(
            Conv2dNormActivation(
                branch2_input,
                branch_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
            ),
            nn.Conv2d(
                branch_channels,
                branch_channels,
                kernel_size=(3, 3),
                stride=(dw_conv_stride, dw_conv_stride),
                padding=(1, 1),
                groups=branch_channels,
                bias=False,
            ),
            nn.BatchNorm2d(branch_channels),
            Conv2dNormActivation(
                branch_channels,
                branch_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.dw_conv_stride == 1:
            (branch1, branch2) = x.chunk(2, dim=1)
            x = torch.concat((branch1, self.branch2(branch2)), dim=1)
        else:
            x = torch.concat((self.branch1(x), self.branch2(x)), dim=1)

        x = channel_shuffle(x, groups=2)
        return x


# pylint: disable=invalid-name
class ShuffleNet_v2(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        stage_repeats = [3, 7, 3]
        out_channels: list[int] = self.config["out_channels"]

        self.stem = Conv2dNormActivation(
            self.input_channels,
            out_channels[0],
            kernel_size=(3, 3),
            stride=(2, 2),
            padding=(1, 1),
            bias=False,
        )

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []

        stages["stage1"] = nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(1, 1))
        return_channels.append(out_channels[0])

        for i, repeat in enumerate(stage_repeats):
            layers = []
            layers.append(ShuffleUnit(in_channels=out_channels[i], out_channels=out_channels[i + 1], dw_conv_stride=2))
            for _ in range(repeat):
                layers.append(
                    ShuffleUnit(in_channels=out_channels[i + 1], out_channels=out_channels[i + 1], dw_conv_stride=1)
                )

            stages[f"stage{i+2}"] = nn.Sequential(*layers)
            return_channels.append(out_channels[i + 1])

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            Conv2dNormActivation(
                out_channels[-2],
                out_channels[-1],
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
            ),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = out_channels[-1]
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config("shufflenet_v2_0_5", ShuffleNet_v2, config={"out_channels": [24, 48, 96, 192, 1024]})
registry.register_model_config("shufflenet_v2_1_0", ShuffleNet_v2, config={"out_channels": [24, 116, 232, 464, 1024]})
registry.register_model_config("shufflenet_v2_1_5", ShuffleNet_v2, config={"out_channels": [24, 176, 352, 704, 1024]})
registry.register_model_config("shufflenet_v2_2_0", ShuffleNet_v2, config={"out_channels": [24, 244, 488, 976, 2048]})

registry.register_weights(
    "shufflenet_v2_1_0_il-common",
    {
        "description": "ShuffleNet v2 1.0x output channels model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 6.4,
                "sha256": "62c435e9a1335aa84290a3c7c99058b0fccd6cc27a04782a8fc1a55e1115f45c",
            }
        },
        "net": {"network": "shufflenet_v2_1_0", "tag": "il-common"},
    },
)



"""
FCMAE, adapted from
https://github.com/facebookresearch/ConvNeXt-V2/blob/main/models/fcmae.py

Paper "ConvNeXt V2: Co-designing and Scaling ConvNets with Masked Autoencoders",
https://arxiv.org/abs/2301.00808
"""

# Reference license: MIT

from typing import Any
from typing import Optional

import torch
from torch import nn

from birder.common.masking import uniform_mask
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.mim.base import MIMBaseNet


class FCMAE(MIMBaseNet):
    default_size = (224, 224)

    def __init__(
        self,
        encoder: PreTrainEncoder,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(encoder, config=config, size=size)
        assert self.config is None, "config not supported"
        assert isinstance(self.encoder, MaskedTokenRetentionMixin)
        assert hasattr(self.encoder, "decoder_block")

        self.mask_ratio = 0.6
        self.decoder_embed_dim = 512
        self.decoder_depth = 1
        self.patch_size = encoder.max_stride

        self.proj = nn.Conv2d(
            in_channels=self.encoder.encoding_size,
            out_channels=self.decoder_embed_dim,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            bias=True,
        )

        self.mask_token = nn.Parameter(torch.zeros(1, self.decoder_embed_dim, 1, 1))

        layers = []
        for _ in range(self.decoder_depth):
            layers.append(self.encoder.decoder_block(self.decoder_embed_dim))

        self.decoder = nn.Sequential(*layers)

        self.pred = nn.Conv2d(
            in_channels=self.decoder_embed_dim,
            out_channels=self.patch_size**2 * self.encoder.input_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            bias=True,
        )

        # Weights initialization
        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.Linear)):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            if isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0)
                nn.init.constant_(m.weight, 1.0)

            if hasattr(self, "mask_token") is True:
                nn.init.normal_(self.mask_token, std=0.02)

    def patchify(self, imgs: torch.Tensor) -> torch.Tensor:
        """
        imgs: (N, 3, H, W)
        x: (N, L, patch_size**2 * 3)
        """

        p = self.patch_size
        assert imgs.shape[2] == imgs.shape[3] and imgs.shape[2] % p == 0

        h = imgs.shape[2] // p
        w = imgs.shape[3] // p
        x = imgs.reshape(shape=(imgs.shape[0], 3, h, p, w, p))
        x = torch.einsum("nchpwq->nhwpqc", x)
        x = x.reshape(shape=(imgs.shape[0], h * w, p**2 * 3))

        return x

    def unpatchify(self, x: torch.Tensor) -> torch.Tensor:
        """
        x: (N, conv_out**2, L*3) or (N, L*3, conv_out, conv_out)
        imgs: (N, 3, H, W)
        """

        if x.ndim == 4:
            (n, c, _, _) = x.shape
            x = x.reshape(n, c, -1)
            x = torch.einsum("ncl->nlc", x)

        p = self.patch_size
        h = int(x.shape[1] ** 0.5)
        w = int(x.shape[1] ** 0.5)
        assert h * w == x.shape[1]

        x = x.reshape(shape=(x.shape[0], h, w, p, p, 3))
        x = torch.einsum("nhwpqc->nchpwq", x)
        imgs = x.reshape(shape=(x.shape[0], 3, h * p, h * p))

        return imgs

    def forward_decoder(self, x: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        x = self.proj(x)

        # Append mask token
        (B, _, H, W) = x.shape
        mask = mask.reshape(-1, H, W).unsqueeze(1).type_as(x)
        mask_token = self.mask_token.repeat(B, 1, H, W)
        x = x * (1.0 - mask) + (mask_token * mask)

        # Decoding
        x = self.decoder(x)
        x = self.pred(x)

        return x

    def forward_loss(self, x: torch.Tensor, pred: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        """
        mask: 0 is keep, 1 is remove
        """

        (n, c, _, _) = pred.shape
        pred = pred.reshape(n, c, -1)
        pred = torch.einsum("ncl->nlc", pred)

        target = self.patchify(x)
        loss = (pred - target) ** 2
        loss = loss.mean(dim=-1)  # [N, L], mean loss per patch
        loss = (loss * mask).sum() / mask.sum()  # Mean loss on removed patches

        return loss

    def forward(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        seq_len = (self.size[0] // self.encoder.max_stride) * (self.size[1] // self.encoder.max_stride)
        mask = uniform_mask(x.size(0), seq_len, mask_ratio=self.mask_ratio, device=x.device)[0]

        latent = self.encoder.masked_encoding_retention(x, mask, return_keys="features")
        pred = self.forward_decoder(latent["features"], mask)
        loss = self.forward_loss(x, pred, mask)

        return {"loss": loss, "pred": pred, "mask": mask}



"""
CrossMAE, adapted from
https://github.com/TonyLianLong/CrossMAE/blob/main/models_cross.py

Paper "Rethinking Patch Dependence for Masked Autoencoders",
https://arxiv.org/abs/2401.14391
"""

# Reference license: Attribution-NonCommercial 4.0 International

from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP

from birder.common.masking import uniform_mask
from birder.net.base import MaskedTokenOmissionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import pos_embedding_sin_cos_2d
from birder.net.mim.base import MIMBaseNet


class WeightedFeatureMaps(nn.Module):
    def __init__(self, k: int, decoder_depth: int) -> None:
        super().__init__()
        self.linear = nn.Linear(k, decoder_depth, bias=False)

    def forward(self, stacked_feature_maps: torch.Tensor) -> torch.Tensor:
        output = self.linear(stacked_feature_maps)

        return output


class CrossAttention(nn.Module):
    def __init__(self, encoder_dim: int, decoder_dim: int, num_heads: int) -> None:
        super().__init__()
        self.num_heads = num_heads
        head_dim = decoder_dim // num_heads
        self.scale = head_dim**-0.5
        self.q = nn.Linear(decoder_dim, decoder_dim)
        self.kv = nn.Linear(encoder_dim, decoder_dim * 2)
        self.proj = nn.Linear(decoder_dim, decoder_dim)

    def forward(self, tgt: torch.Tensor, memory: torch.Tensor) -> torch.Tensor:
        (B, N, C) = tgt.size()
        n_kv = memory.size(1)
        q = self.q(tgt).reshape(B, N, self.num_heads, C // self.num_heads).permute(0, 2, 1, 3)
        kv = self.kv(memory).reshape(B, n_kv, 2, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        (k, v) = kv.unbind(0)

        attn = F.scaled_dot_product_attention(q, k, v, dropout_p=0.0)  # pylint: disable=not-callable
        x = attn.transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)

        return x


class CrossAttentionBlock(nn.Module):
    def __init__(self, encoder_dim: int, decoder_dim: int, num_heads: int, mlp_ratio: float) -> None:
        super().__init__()
        self.norm1 = nn.LayerNorm(decoder_dim, eps=1e-6)
        self.cross_attn = CrossAttention(encoder_dim, decoder_dim, num_heads=num_heads)
        self.norm2 = nn.LayerNorm(decoder_dim, eps=1e-6)
        self.mlp = MLP(decoder_dim, [int(decoder_dim * mlp_ratio), decoder_dim], activation_layer=nn.GELU)

    def forward(self, tgt: torch.Tensor, memory: torch.Tensor) -> torch.Tensor:
        x = tgt + self.cross_attn(self.norm1(tgt), memory)
        x = x + self.mlp(self.norm2(x))
        return x


class CrossMAE(MIMBaseNet):
    default_size = (224, 224)

    def __init__(
        self,
        encoder: PreTrainEncoder,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(encoder, config=config, size=size)
        assert self.config is None, "config not supported"
        assert isinstance(self.encoder, MaskedTokenOmissionMixin)

        self.mask_ratio = 0.75
        self.kept_mask_ratio = 0.25
        self.patch_size = self.encoder.stem_stride
        encoder_dim = self.encoder.embedding_size
        decoder_embed_dim = 512
        decoder_depth = 8

        self.wfm = WeightedFeatureMaps(self.encoder.num_layers, decoder_depth=decoder_depth)
        self.decoder_norms = nn.ModuleList()
        for _ in range(decoder_depth - 1):
            self.decoder_norms.append(nn.LayerNorm(encoder_dim, eps=1e-6))

        self.decoder_norms.append(nn.Identity())  # Last norm uses the encoder native norm

        self.mask_token = nn.Parameter(torch.zeros(1, 1, decoder_embed_dim))

        # Fixed sin-cos embedding
        pos_embedding = pos_embedding_sin_cos_2d(
            h=self.size[0] // self.patch_size,
            w=self.size[1] // self.patch_size,
            dim=decoder_embed_dim,
            num_special_tokens=0,
        ).unsqueeze(0)
        self.decoder_pos_embed = nn.Parameter(pos_embedding, requires_grad=False)

        self.decoder_layers = nn.ModuleList()
        for _ in range(decoder_depth):
            self.decoder_layers.append(CrossAttentionBlock(encoder_dim, decoder_embed_dim, num_heads=16, mlp_ratio=4.0))

        self.decoder_norm = nn.LayerNorm(decoder_embed_dim, eps=1e-6)
        self.pred = nn.Linear(decoder_embed_dim, self.patch_size**2 * self.encoder.input_channels, bias=True)

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.LayerNorm):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def patchify(self, imgs: torch.Tensor) -> torch.Tensor:
        """
        imgs: (N, 3, H, W)
        x: (N, L, patch_size**2 * 3)
        """

        p = self.patch_size
        assert imgs.shape[2] == imgs.shape[3] and imgs.shape[2] % p == 0

        h = imgs.shape[2] // p
        w = imgs.shape[3] // p
        x = imgs.reshape(shape=(imgs.shape[0], 3, h, p, w, p))
        x = torch.einsum("nchpwq->nhwpqc", x)
        x = x.reshape(shape=(imgs.shape[0], h * w, p**2 * 3))

        return x

    def unpatchify(self, x: torch.Tensor) -> torch.Tensor:
        """
        x: (N, L, patch_size**2 * 3)
        imgs: (N, 3, H, W)
        """

        p = self.patch_size
        h = int(x.shape[1] ** 0.5)
        w = int(x.shape[1] ** 0.5)
        assert h * w == x.shape[1]

        x = x.reshape(shape=(x.shape[0], h, w, p, p, 3))
        x = torch.einsum("nhwpqc->nchpwq", x)
        imgs = x.reshape(shape=(x.shape[0], 3, h * p, h * p))

        return imgs

    def fill_pred(self, mask: torch.Tensor, pred: torch.Tensor) -> torch.Tensor:
        (N, L) = mask.shape[0:2]
        combined = torch.zeros(N, L, pred.shape[2], device=pred.device, dtype=pred.dtype)
        combined[mask.bool()] = pred.view(-1, pred.shape[2])

        return combined

    def mask_tokens_grid(self, mask: torch.Tensor) -> torch.Tensor:
        N = mask.size(0)
        x = self.decoder_pos_embed.masked_select(mask.bool().unsqueeze(-1)).reshape(N, -1, self.mask_token.size(-1))
        x = x + self.mask_token

        return x

    def forward_decoder(self, memory: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        x = self.mask_tokens_grid(mask)
        memory = self.wfm(memory)

        for i, layer in enumerate(self.decoder_layers):
            x = layer(x, self.decoder_norms[i](memory[..., i]))

        x = self.decoder_norm(x)
        x = self.pred(x)

        return x

    def forward_loss(self, x: torch.Tensor, pred: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        target = self.patchify(x)

        target = target.masked_select(mask.bool().unsqueeze(-1)).reshape(target.size(0), -1, target.size(-1))

        # Normalize pixels per-patch
        mean = target.mean(dim=-1, keepdim=True)
        var = target.var(dim=-1, keepdim=True)
        target = (target - mean) / (var + 1.0e-6) ** 0.5

        loss = (pred - target) ** 2
        loss = loss.mean()

        return loss

    def forward(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        seq_len = (self.size[0] // self.encoder.stem_stride) * (self.size[1] // self.encoder.stem_stride)
        (mask, ids_keep, _) = uniform_mask(x.size(0), seq_len, self.mask_ratio, self.kept_mask_ratio, device=x.device)

        latent = self.encoder.masked_encoding_omission(x, ids_keep, return_all_features=True)["tokens"]
        pred = self.forward_decoder(latent, mask)
        loss = self.forward_loss(x, pred, mask)

        return {"loss": loss, "pred": pred, "mask": mask}



"""
DaViT, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/davit.py

Paper "DaViT: Dual Attention Vision Transformers", https://arxiv.org/abs/2204.03645

Changes from original:
* Window size based on image size (image size // 32)
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.layers import LayerNorm2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType


def window_partition(x: torch.Tensor, window_size: tuple[int, int]) -> torch.Tensor:
    (B, H, W, C) = x.shape
    x = x.view(B, H // window_size[0], window_size[0], W // window_size[1], window_size[1], C)
    windows = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size[0], window_size[1], C)

    return windows


class Stem(nn.Module):
    def __init__(self, in_channels: int, out_channels: int):
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=(7, 7),
            stride=(4, 4),
            padding=(3, 3),
        )
        self.norm = LayerNorm2d(out_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        x = self.norm(x)
        return x


class ConvPosEnc(nn.Module):
    def __init__(self, dim: int, kernel_size: tuple[int, int], act: bool) -> None:
        super().__init__()
        self.proj = nn.Conv2d(
            dim,
            dim,
            kernel_size=kernel_size,
            stride=(1, 1),
            padding=(kernel_size[0] // 2, kernel_size[1] // 2),
            groups=dim,
        )
        if act is True:
            self.act = nn.GELU()
        else:
            self.act = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.act(self.proj(x))
        return x


class Downsample(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int) -> None:
        super().__init__()
        self.norm = LayerNorm2d(in_channels)
        self.even_k = kernel_size % 2 == 0
        if self.even_k is True:
            padding = (0, 0)
        else:
            padding = (kernel_size // 2, kernel_size // 2)

        self.conv = nn.Conv2d(
            in_channels, out_channels, kernel_size=(kernel_size, kernel_size), stride=(2, 2), padding=padding
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (_, _, H, W) = x.shape
        x = self.norm(x)
        if self.even_k is True:
            (k_h, k_w) = self.conv.kernel_size
            pad_r = (k_w - W % k_w) % k_w
            pad_b = (k_h - H % k_h) % k_h
            x = F.pad(x, (0, pad_r, 0, pad_b))

        x = self.conv(x)
        return x


class ChannelAttention(nn.Module):
    def __init__(self, dim: int, num_heads: int, qkv_bias: bool) -> None:
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, N, C) = x.shape

        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        (q, k, v) = qkv.unbind(0)

        k = k * self.scale
        attn = k.transpose(-1, -2) @ v
        attn = attn.softmax(dim=-1)
        x = (attn @ q.transpose(-1, -2)).transpose(-1, -2)
        x = x.transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)

        return x


class ChannelBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float,
        qkv_bias: bool,
        drop_path: float,
        cpe_act: bool = False,
    ) -> None:
        super().__init__()
        self.cpe1 = ConvPosEnc(dim=dim, kernel_size=(3, 3), act=cpe_act)
        self.norm1 = nn.LayerNorm(dim)
        self.attn = ChannelAttention(dim, num_heads=num_heads, qkv_bias=qkv_bias)

        self.cpe2 = ConvPosEnc(dim=dim, kernel_size=(3, 3), act=cpe_act)
        self.norm2 = nn.LayerNorm(dim)
        self.mlp = MLP(dim, [int(dim * mlp_ratio), dim], activation_layer=nn.GELU)
        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, C, H, W) = x.shape
        x = self.cpe1(x).flatten(2).transpose(1, 2)

        cur = self.norm1(x)
        cur = self.attn(cur)
        x = x + self.drop_path(cur)

        x = self.cpe2(x.transpose(1, 2).view(B, C, H, W))
        x = x.flatten(2).transpose(1, 2)
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        x = x.transpose(1, 2).view(B, C, H, W)

        return x


class WindowAttention(nn.Module):
    def __init__(self, dim: int, num_heads: int, qkv_bias: bool) -> None:
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, N, C) = x.shape

        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        (q, k, v) = qkv.unbind(0)

        x = F.scaled_dot_product_attention(q, k, v, scale=self.scale)  # pylint: disable=not-callable
        x = x.transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)

        return x


class SpatialBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        window_size: tuple[int, int],
        mlp_ratio: float,
        qkv_bias: bool,
        drop_path: float,
        cpe_act: bool,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.window_size = window_size

        self.cpe1 = ConvPosEnc(dim=dim, kernel_size=(3, 3), act=cpe_act)
        self.norm1 = nn.LayerNorm(dim)
        self.attn = WindowAttention(dim, num_heads=num_heads, qkv_bias=qkv_bias)

        self.cpe2 = ConvPosEnc(dim=dim, kernel_size=(3, 3), act=cpe_act)
        self.norm2 = nn.LayerNorm(dim)
        self.mlp = MLP(dim, [int(dim * mlp_ratio), dim], activation_layer=nn.GELU)
        self.drop_path = StochasticDepth(drop_path, mode="row")

    # pylint: disable=invalid-name
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, C, H, W) = x.shape

        shortcut = self.cpe1(x).flatten(2).transpose(1, 2)

        x = self.norm1(shortcut)
        x = x.view(B, H, W, C)

        pad_l = pad_t = 0
        pad_r = (self.window_size[1] - W % self.window_size[1]) % self.window_size[1]
        pad_b = (self.window_size[0] - H % self.window_size[0]) % self.window_size[0]
        x = F.pad(x, (0, 0, pad_l, pad_r, pad_t, pad_b))
        (_, Hp, Wp, _) = x.shape

        x_windows = window_partition(x, self.window_size)
        x_windows = x_windows.view(-1, self.window_size[0] * self.window_size[1], C)

        # W-MSA/SW-MSA
        attn_windows = self.attn(x_windows)

        # Merge windows
        attn_windows = attn_windows.view(-1, self.window_size[0], self.window_size[1], C)
        Ca = attn_windows.shape[-1]
        x = attn_windows.view(
            -1, Hp // self.window_size[0], Wp // self.window_size[1], self.window_size[0], self.window_size[1], Ca
        )
        x = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, Hp, Wp, Ca)

        # In case pad_r > 0 or pad_b > 0
        x = x[:, :H, :W, :].contiguous()

        x = x.view(B, H * W, C)
        x = shortcut + self.drop_path(x)

        x = self.cpe2(x.transpose(1, 2).view(B, C, H, W))
        x = x.flatten(2).transpose(1, 2)
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        x = x.transpose(1, 2).view(B, C, H, W)

        return x


class DaViTStage(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        depth: int,
        downsample: bool,
        num_heads: int,
        window_size: tuple[int, int],
        mlp_ratio: float,
        qkv_bias: bool,
        drop_path_rates: list[float],
        cpe_act: bool,
        down_kernel_size: int,
    ) -> None:
        super().__init__()
        if downsample:
            self.downsample = Downsample(in_channels, out_channels, kernel_size=down_kernel_size)
        else:
            self.downsample = nn.Identity()

        stage_blocks = []
        for block_idx in range(depth):
            stage_blocks.append(
                SpatialBlock(
                    out_channels,
                    num_heads=num_heads,
                    window_size=window_size,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    drop_path=drop_path_rates[block_idx],
                    cpe_act=cpe_act,
                )
            )
            stage_blocks.append(
                ChannelBlock(
                    out_channels,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    drop_path=drop_path_rates[block_idx],
                    cpe_act=cpe_act,
                )
            )

        self.blocks = nn.Sequential(*stage_blocks)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = self.blocks(x)
        return x


class DaViT(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.blocks.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        mlp_ratio = 4.0
        qkv_bias = True
        down_kernel_size = 2
        window_size = (int(self.size[0] / (2**5)), int(self.size[1] / (2**5)))
        depths: list[int] = self.config["depths"]
        dims: list[int] = self.config["dims"]
        heads: list[int] = self.config["heads"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.stem = Stem(self.input_channels, dims[0])

        num_stages = len(depths)
        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        in_channels = dims[0]
        for stage_idx in range(num_stages):
            out_channels = dims[stage_idx]
            stages[f"stage{stage_idx+1}"] = DaViTStage(
                in_channels,
                out_channels,
                depth=depths[stage_idx],
                downsample=stage_idx > 0,
                num_heads=heads[stage_idx],
                window_size=window_size,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                drop_path_rates=dpr[stage_idx],
                cpe_act=False,
                down_kernel_size=down_kernel_size,
            )
            return_channels.append(out_channels)
            in_channels = out_channels

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            LayerNorm2d(dims[-1], eps=1e-6),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = dims[-1]
        self.classifier = self.create_classifier()

        self.stem_stride = 4
        self.stem_width = dims[0]
        self.encoding_size = dims[-1]

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:  # pylint:disable=useless-parent-delegation
        super().set_dynamic_size(dynamic_size)
        # No action taken, just run with constant window form initialization

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        super().adjust_size(new_size)

        new_window_size = (int(new_size[0] / (2**5)), int(new_size[1] / (2**5)))
        for m in self.body.modules():
            if isinstance(m, SpatialBlock):
                m.window_size = new_window_size


registry.register_model_config(
    "davit_tiny",
    DaViT,
    config={"depths": [1, 1, 3, 1], "dims": [96, 192, 384, 768], "heads": [3, 6, 12, 24], "drop_path_rate": 0.1},
)
registry.register_model_config(
    "davit_small",
    DaViT,
    config={"depths": [1, 1, 9, 1], "dims": [96, 192, 384, 768], "heads": [3, 6, 12, 24], "drop_path_rate": 0.2},
)
registry.register_model_config(
    "davit_base",
    DaViT,
    config={"depths": [1, 1, 9, 1], "dims": [128, 256, 512, 1024], "heads": [4, 8, 16, 32], "drop_path_rate": 0.4},
)
registry.register_model_config(
    "davit_large",
    DaViT,
    config={"depths": [1, 1, 9, 1], "dims": [192, 384, 768, 1536], "heads": [6, 12, 24, 48], "drop_path_rate": 0.4},
)
registry.register_model_config(
    "davit_huge",
    DaViT,
    config={"depths": [1, 1, 9, 1], "dims": [256, 512, 1024, 2048], "heads": [8, 16, 32, 64], "drop_path_rate": 0.5},
)
registry.register_model_config(
    "davit_giant",
    DaViT,
    config={"depths": [1, 1, 12, 3], "dims": [384, 768, 1536, 3072], "heads": [12, 24, 48, 96], "drop_path_rate": 0.5},
)

registry.register_weights(
    "davit_tiny_il-all",
    {
        "url": "https://huggingface.co/birder-project/davit_tiny_il-all/resolve/main",
        "description": "DaViT tiny model trained on the il-all dataset",
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 107.0,
                "sha256": "ba42050975d56527b61eac6faf0de77bb6aa0467fd676ef3556654cd967146a4",
            }
        },
        "net": {"network": "davit_tiny", "tag": "il-all"},
    },
)



"""
Scale-Aware Modulation Transformer (SMT), adapted from
https://github.com/AFeng-x/SMT/blob/main/models/smt.py

Paper "Scale-Aware Modulation Meet Transformer", https://arxiv.org/abs/2307.08579
"""

# Reference license: MIT

import math
from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import StochasticDepth

from birder.layers import LayerScale
from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class SequentialWithShape(nn.Sequential):
    def forward(self, x: torch.Tensor, H: int, W: int) -> torch.Tensor:  # pylint: disable=arguments-differ
        for module in self:
            x = module(x, H, W)

        return x


class DWConv(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.dwconv = nn.Conv2d(dim, dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim)

    def forward(self, x: torch.Tensor, H: int, W: int) -> torch.Tensor:
        (B, _, C) = x.size()
        x = x.transpose(1, 2).view(B, C, H, W)
        x = self.dwconv(x)
        x = x.flatten(2).transpose(1, 2)

        return x


class MLP(nn.Module):
    def __init__(self, in_features: int, hidden_features: int) -> None:
        super().__init__()
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.dwconv = DWConv(hidden_features)
        self.act = nn.GELU()
        self.fc2 = nn.Linear(hidden_features, in_features)

    def forward(self, x: torch.Tensor, H: int, W: int) -> torch.Tensor:
        x = self.fc1(x)
        x = self.act(x + self.dwconv(x, H, W))
        x = self.fc2(x)

        return x


class CAAttention(nn.Module):
    def __init__(self, dim: int, ca_num_heads: int, qkv_bias: bool, proj_drop: float, expand_ratio: int) -> None:
        super().__init__()
        assert dim % ca_num_heads == 0, f"dim {dim} should be divided by num_heads {ca_num_heads}"
        self.ca_num_heads = ca_num_heads

        self.act = nn.GELU()
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.split_groups = dim // ca_num_heads

        self.v = nn.Linear(dim, dim, bias=qkv_bias)
        self.s = nn.Linear(dim, dim, bias=qkv_bias)
        self.local_conv = nn.ModuleList()
        for i in range(self.ca_num_heads):
            self.local_conv.append(
                nn.Conv2d(
                    dim // self.ca_num_heads,
                    dim // self.ca_num_heads,
                    kernel_size=(3 + i * 2, 3 + i * 2),
                    stride=(1, 1),
                    padding=(1 + i, 1 + i),
                    groups=dim // self.ca_num_heads,
                )
            )

        self.proj0 = nn.Conv2d(
            dim, dim * expand_ratio, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), groups=self.split_groups
        )
        self.bn = nn.BatchNorm2d(dim * expand_ratio)
        self.proj1 = nn.Conv2d(dim * expand_ratio, dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0))

    def forward(self, x: torch.Tensor, H: int, W: int) -> torch.Tensor:
        (B, N, C) = x.size()

        v = self.v(x)
        s = self.s(x).reshape(B, H, W, self.ca_num_heads, C // self.ca_num_heads).permute(3, 0, 4, 1, 2)
        s_i_out = []
        for i, local_conv in enumerate(self.local_conv):
            s_i = s[i]
            s_i = local_conv(s_i).reshape(B, self.split_groups, -1, H, W)
            s_i_out.append(s_i)

        s_out = torch.concat(s_i_out, dim=2)
        s_out = s_out.reshape(B, C, H, W)
        s_out = self.proj1(self.act(self.bn(self.proj0(s_out))))
        s_out = s_out.reshape(B, C, N).permute(0, 2, 1)
        x = s_out * v

        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class SAAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        sa_num_heads: int,
        qkv_bias: bool,
        proj_drop: float,
        attn_drop: float,
    ) -> None:
        super().__init__()
        assert dim % sa_num_heads == 0, f"dim {dim} should be divided by num_heads {sa_num_heads}"
        self.sa_num_heads = sa_num_heads

        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        head_dim = dim // sa_num_heads
        self.scale = head_dim**-0.5
        self.q = nn.Linear(dim, dim, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.kv = nn.Linear(dim, dim * 2, bias=qkv_bias)
        self.conv = nn.Conv2d(dim, dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim)

    def forward(self, x: torch.Tensor, H: int, W: int) -> torch.Tensor:
        (B, N, C) = x.size()

        q = self.q(x).reshape(B, N, self.sa_num_heads, C // self.sa_num_heads).permute(0, 2, 1, 3)
        kv = self.kv(x).reshape(B, -1, 2, self.sa_num_heads, C // self.sa_num_heads).permute(2, 0, 3, 1, 4)
        (k, v) = kv.unbind(0)
        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        x = (attn @ v).transpose(1, 2).reshape(B, N, C) + self.conv(
            v.transpose(1, 2).reshape(B, N, C).transpose(1, 2).view(B, C, H, W)
        ).view(B, C, N).transpose(1, 2)

        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class Attention(nn.Module):
    def __init__(
        self,
        dim: int,
        ca_num_heads: int,
        sa_num_heads: int,
        qkv_bias: bool,
        proj_drop: float,
        attn_drop: float,
        ca_attention: bool,
        expand_ratio: int,
    ) -> None:
        super().__init__()
        if ca_attention is True:
            self.attn = CAAttention(
                dim, ca_num_heads=ca_num_heads, qkv_bias=qkv_bias, proj_drop=proj_drop, expand_ratio=expand_ratio
            )
        else:
            self.attn = SAAttention(
                dim, sa_num_heads=sa_num_heads, qkv_bias=qkv_bias, proj_drop=proj_drop, attn_drop=attn_drop
            )

    def forward(self, x: torch.Tensor, H: int, W: int) -> torch.Tensor:
        return self.attn(x, H, W)


class SMTBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        ca_num_heads: int,
        sa_num_heads: int,
        mlp_ratio: float,
        qkv_bias: bool,
        layer_scale_value: Optional[float],
        proj_drop: float,
        attn_drop: float,
        drop_path: float,
        ca_attention: bool,
        expand_ratio: int,
    ) -> None:
        super().__init__()
        self.norm1 = nn.LayerNorm(dim, eps=1e-6)
        self.attn = Attention(
            dim,
            ca_num_heads=ca_num_heads,
            sa_num_heads=sa_num_heads,
            qkv_bias=qkv_bias,
            attn_drop=attn_drop,
            proj_drop=proj_drop,
            ca_attention=ca_attention,
            expand_ratio=expand_ratio,
        )
        self.drop_path = StochasticDepth(drop_path, mode="row")
        self.norm2 = nn.LayerNorm(dim, eps=1e-6)
        self.mlp = MLP(dim, int(mlp_ratio * dim))

        if layer_scale_value is not None:
            self.layer_scale_1 = LayerScale(dim, layer_scale_value)
            self.layer_scale_2 = LayerScale(dim, layer_scale_value)
        else:
            self.layer_scale_1 = nn.Identity()
            self.layer_scale_2 = nn.Identity()

    def forward(self, x: torch.Tensor, H: int, W: int) -> torch.Tensor:
        x = x + self.drop_path(self.layer_scale_1(self.attn(self.norm1(x), H, W)))
        x = x + self.drop_path(self.layer_scale_2(self.mlp(self.norm2(x), H, W)))

        return x


class OverlapPatchEmbed(nn.Module):
    def __init__(self, patch_size: tuple[int, int], stride: tuple[int, int], in_channels: int, embed_dim: int) -> None:
        super().__init__()
        self.proj = nn.Conv2d(
            in_channels,
            embed_dim,
            kernel_size=patch_size,
            stride=stride,
            padding=(patch_size[0] // 2, patch_size[1] // 2),
        )
        self.norm = nn.LayerNorm(embed_dim)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, int, int]:
        x = self.proj(x)
        (_, _, H, W) = x.size()
        x = x.flatten(2).transpose(1, 2)
        x = self.norm(x)

        return (x, H, W)


class Stem(nn.Module):
    def __init__(self, kernel_size: tuple[int, int], stride: tuple[int, int], in_channels: int, embed_dim: int) -> None:
        super().__init__()
        self.conv = nn.Sequential(
            Conv2dNormActivation(
                in_channels,
                embed_dim,
                kernel_size=kernel_size,
                stride=stride,
                padding=(kernel_size[0] // 2, kernel_size[1] // 2),
            ),
            nn.Conv2d(embed_dim, embed_dim, kernel_size=(2, 2), stride=(2, 2), padding=(0, 0)),
        )
        self.norm = nn.LayerNorm(embed_dim)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, int, int]:
        x = self.conv(x)
        (_, _, H, W) = x.size()
        x = x.flatten(2).transpose(1, 2)
        x = self.norm(x)

        return (x, H, W)


class SMTStage(nn.Module):
    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(
        self,
        dim: int,
        dim_out: int,
        depth: int,
        downsample_stem: bool,
        stem_conv: tuple[int, int],
        ca_num_heads: int,
        sa_num_heads: int,
        mlp_ratio: float,
        qkv_bias: bool,
        layer_scale_value: Optional[float],
        proj_drop: float,
        attn_drop: float,
        drop_path: list[float],
        ca_attention: Optional[bool],
        expand_ratio: int,
    ) -> None:
        super().__init__()
        if downsample_stem is True:
            self.downsample_block = Stem(kernel_size=stem_conv, stride=(2, 2), in_channels=dim, embed_dim=dim_out)
        else:
            self.downsample_block = OverlapPatchEmbed(
                patch_size=(3, 3), stride=(2, 2), in_channels=dim, embed_dim=dim_out
            )

        layers = []
        for i in range(depth):
            if ca_attention is not None:
                caa = ca_attention
            else:
                caa = i % 2 == 0

            layers.append(
                SMTBlock(
                    dim_out,
                    ca_num_heads=ca_num_heads,
                    sa_num_heads=sa_num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    layer_scale_value=layer_scale_value,
                    proj_drop=proj_drop,
                    attn_drop=attn_drop,
                    drop_path=drop_path[i],
                    ca_attention=caa,
                    expand_ratio=expand_ratio,
                )
            )

        self.blocks = SequentialWithShape(*layers)
        self.norm = nn.LayerNorm(dim_out, eps=1e-6)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        B = x.size(0)
        (x, H, W) = self.downsample_block(x)
        x = self.blocks(x, H, W)
        x = self.norm(x)
        x = x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()

        return x


class SMT(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        qkv_bias = True
        ca_num_heads = [4, 4, 4, -1]
        sa_num_heads = [-1, -1, 8, 16]
        ca_attentions = [True, True, None, False]  # Tri-state: None activated on even blocks only
        expand_ratio = 2
        embed_dims: list[int] = self.config["embed_dims"]
        mlp_ratios: list[float] = self.config["mlp_ratios"]
        depths: list[int] = self.config["depths"]
        stem_conv: tuple[int, int] = self.config["stem_conv"]
        drop_path_rate: float = self.config["drop_path_rate"]

        num_stages = len(depths)
        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]
        prev_dim = self.input_channels
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(num_stages):
            stages[f"stage{i+1}"] = SMTStage(
                prev_dim,
                embed_dims[i],
                depth=depths[i],
                downsample_stem=i == 0,
                stem_conv=stem_conv,
                ca_num_heads=ca_num_heads[i],
                sa_num_heads=sa_num_heads[i],
                mlp_ratio=mlp_ratios[i],
                qkv_bias=qkv_bias,
                layer_scale_value=None,
                proj_drop=0.0,
                attn_drop=0.0,
                drop_path=dpr[i],
                ca_attention=ca_attentions[i],
                expand_ratio=expand_ratio,
            )
            return_channels.append(embed_dims[i])
            prev_dim = embed_dims[i]

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dims[-1]
        self.classifier = self.create_classifier()

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.Conv2d):
                fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                fan_out //= m.groups
                nn.init.normal_(m.weight, mean=0.0, std=math.sqrt(2.0 / fan_out))
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.LayerNorm):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "smt_t",
    SMT,
    config={
        "embed_dims": [64, 128, 256, 512],
        "mlp_ratios": [4.0, 4.0, 4.0, 2.0],
        "depths": [2, 2, 8, 1],
        "stem_conv": (3, 3),
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "smt_s",
    SMT,
    config={
        "embed_dims": [64, 128, 256, 512],
        "mlp_ratios": [4.0, 4.0, 4.0, 2.0],
        "depths": [3, 4, 18, 2],
        "stem_conv": (3, 3),
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "smt_b",
    SMT,
    config={
        "embed_dims": [64, 128, 256, 512],
        "mlp_ratios": [8.0, 6.0, 4.0, 2.0],
        "depths": [4, 6, 28, 2],
        "stem_conv": (7, 7),
        "drop_path_rate": 0.3,
    },
)
registry.register_model_config(
    "smt_l",
    SMT,
    config={
        "embed_dims": [96, 192, 384, 768],
        "mlp_ratios": [8.0, 6.0, 4.0, 2.0],
        "depths": [4, 6, 28, 4],
        "stem_conv": (7, 7),
        "drop_path_rate": 0.5,
    },
)

registry.register_weights(
    "smt_t_il-common",
    {
        "description": "SMT tiny model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 42.8,
                "sha256": "5b6d2541ed382d85bda88cf19ebf2d8fb0adb1b6475209f17eacf4a2f428bcc0",
            }
        },
        "net": {"network": "smt_t", "tag": "il-common"},
    },
)
